#
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import json
import re
from typing import List, Dict, Any
from collections import defaultdict

from knext.schema.model.schema_helper import SPGTypeName
from knext.builder.operator.spg_record import SPGRecord
from knext.builder.operator.builtin.auto_prompt import AutoPrompt


class OneKEPrompt(AutoPrompt):
    template_zh: str = ""
    template_en: str = ""

    def __init__(self, **kwargs):
        types_list = kwargs.get("types_list", [])
        language = kwargs.get("language", "zh")
        with_description = kwargs.get("language", False)
        split_num = kwargs.get("split_num", 4)
        super().__init__(types_list)
        if language == "zh":
            self.template = self.template_zh
        else:
            self.template = self.template_en
        self.with_description = with_description
        self.split_num = split_num

        self._init_render_variables()
        self._render()

        self.params = kwargs

    def build_prompt(self, variables: Dict[str, str]) -> List[str]:
        instructions = []
        for schema in self.schema_list:
            instructions.append(
                json.dumps(
                    {
                        "instruction": self.template,
                        "schema": schema,
                        "input": variables.get("input"),
                    },
                    ensure_ascii=False,
                )
            )
        return instructions

    def parse_response(self, response: str) -> List[SPGRecord]:
        raise NotImplementedError

    def _render(self):
        raise NotImplementedError

    def multischema_split_by_num(self, split_num, schemas: List[Any]):
        negative_length = max(len(schemas) // split_num, 1) * split_num
        total_schemas = []
        for i in range(0, negative_length, split_num):
            total_schemas.append(schemas[i : i + split_num])

        remain_len = max(1, split_num // 2)
        tmp_schemas = schemas[negative_length:]
        if len(schemas) - negative_length >= remain_len and len(tmp_schemas) > 0:
            total_schemas.append(tmp_schemas)
        elif len(tmp_schemas) > 0:
            total_schemas[-1].extend(tmp_schemas)
        return total_schemas


class OneKE_NERPrompt(OneKEPrompt):
    template_zh: str = (
        "你是专门进行实体抽取的专家。请从input中抽取出符合schema定义的实体，不存在的实体类型返回空列表。请按照JSON字符串的格式回答。"
    )
    template_en: str = "You are an expert in named entity recognition. Please extract entities that match the schema definition from the input. Return an empty list if the entity type does not exist. Please respond in the format of a JSON string."

    def __init__(
        self,
        entity_types: List[SPGTypeName],
        language: str = "zh",
        with_description: bool = False,
        split_num: int = 4,
    ):
        super().__init__(
            types_list=entity_types,
            language=language,
            with_description=with_description,
            split_num=split_num,
        )

    def parse_response(self, response: str) -> List[SPGRecord]:
        if isinstance(response, list) and len(response) > 0:
            response = response[0]
        try:
            ent_obj = json.loads(response)
        except json.decoder.JSONDecodeError:
            print("DeepKE_NERPrompt response JSONDecodeError error.")
            return []
        if type(ent_obj) != dict:
            print("DeepKE_NERPrompt response type error.")
            return []

        spg_records = []
        for type_zh, values in ent_obj.items():
            if type_zh not in self.spg_type_schema_info_zh:
                print(f"Unrecognized entity_type: {type_zh}")
                continue
            type_en, _ = self.spg_type_schema_info_zh[type_zh]
            spg_record = SPGRecord(type_en)
            for value in values:
                spg_record.upsert_properties({"id": value, "name": value})
            spg_records.append(spg_record)

    def _render(self):
        entity_list = []
        for spg_type in self.spg_types:
            entity_list.append(spg_type.name_zh)
        self.schema_list = self.multischema_split_by_num(self.split_num, entity_list)


class OneKE_SPOPrompt(OneKEPrompt):
    template_zh: str = (
        "你是专门进行SPO三元组抽取的专家。请从input中抽取出符合schema定义的spo关系三元组，不存在的关系返回空列表。请按照JSON字符串的格式回答。"
    )
    template_en: str = "You are an expert in spo(subject, predicate, object) triples extraction. Please extract SPO relationship triples that match the schema definition from the input. Return an empty list for relationships that do not exist. Please respond in the format of a JSON string."

    def __init__(
        self,
        spo_types: List[SPGTypeName],
        language: str = "zh",
        with_description: bool = False,
        split_num: int = 4,
    ):
        super().__init__(
            types_list=spo_types,
            language=language,
            with_description=with_description,
            split_num=split_num,
        )
        self.properties_mapper = {}
        self.relations_mapper = {}

    def parse_response(self, response: str) -> List[SPGRecord]:
        if isinstance(response, list) and len(response) > 0:
            response = response[0]
        try:
            re_obj = json.loads(response)
        except json.decoder.JSONDecodeError:
            print("DeepKE_REPrompt response JSONDecodeError error.")
            return []
        if type(re_obj) != dict:
            print("DeepKE_REPrompt response type error.")
            return []

        relation_dcir = defaultdict(list)
        for relation_zh, values in re_obj.items():
            if relation_zh not in self.property_info_zh[relation_zh]:
                print(f"Unrecognized relation: {relation_zh}")
                continue
            if values and isinstance(values, list):
                for value in values:
                    if (
                        type(value) != dict
                        or "subject" not in value
                        or "object" not in value
                    ):
                        print("DeepKE_REPrompt response type error.")
                        continue
                    s_zh, o_zh = value.get("subject", ""), value.get("object", "")
                    relation_dcir[relation_zh].append((s_zh, o_zh))

        spg_records = []
        for relation_zh, sub_obj_list in relation_dcir.items():
            sub_dict = defaultdict(list)
            for s_zh, o_zh in sub_obj_list:
                sub_dict[s_zh].append(o_zh)
            for s_zh, o_list in sub_dict.items():
                if s_zh in self.spg_type_schema_info_zh:
                    print(f"Unrecognized subject_type: {s_zh}")
                    continue
                object_value = ",".join(o_list)
                s_type_zh = self.properties_mapper.get(relation_zh, None)
                if s_type_zh is not None:
                    s_type_en, _ = self.spg_type_schema_info_zh[s_type_zh]
                    relation_en, _ = self.property_info_zh[relation_zh]
                    spg_record = SPGRecord(s_type_en).upsert_properties(
                        {"id": s_zh, "name": s_zh}
                    )
                    spg_record.upsert_property(relation_en, object_value)
                else:
                    s_type_zh, o_type_zh = self.relations_mapper.get(
                        relation_zh, [None, None]
                    )
                    if s_type_zh is None or o_type_zh is None:
                        print(f"Unrecognized relation: {relation_zh}")
                        continue
                    s_type_en, _ = self.spg_type_schema_info_zh[s_type_zh]
                    spg_record = SPGRecord(s_type_en).upsert_properties(
                        {"id": s_zh, "name": s_zh}
                    )
                    relation_en, _, object_type = self.relation_info_zh[s_type_zh][
                        relation_zh
                    ]
                    spg_record.upsert_relation(relation_en, object_type, object_value)
                spg_records.append(spg_record)
        return spg_records

    def _render(self):
        spo_list = []
        for spg_type in self.spg_types:
            type_en, _ = self.spg_type_schema_info_zh[spg_type]
            for v in spg_type.properties.values():
                spo_list.append(
                    {
                        "subject_type": spg_type.name_zh,
                        "predicate": v.name_zh,
                        "object_type": "文本",
                    }
                )
                self.properties_mapper[v.name_zh] = spg_type
            for v in spg_type.relations.values():
                _, _, object_type = self.relation_info_en[type_en][v.name]
                spo_list.append(
                    {
                        "subject_type": spg_type.name_zh,
                        "predicate": v.name_zh,
                        "object_type": object_type,
                    }
                )
                self.relations_mapper[v.name_zh] = [spg_type, object_type]
        self.schema_list = self.multischema_split_by_num(self.split_num, spo_list)


class OneKE_REPrompt(OneKE_SPOPrompt):
    template_zh: str = (
        "你是专门进行关系抽取的专家。请从input中抽取出符合schema定义的关系三元组，不存在的关系返回空列表。请按照JSON字符串的格式回答。"
    )
    template_en: str = "You are an expert in relationship extraction. Please extract relationship triples that match the schema definition from the input. Return an empty list for relationships that do not exist. Please respond in the format of a JSON string."

    def __init__(
        self,
        relation_types: List[SPGTypeName],
        language: str = "zh",
        with_description: bool = False,
        split_num: int = 4,
    ):
        super().__init__(relation_types, language, with_description, split_num)

    def _render(self):
        re_list = []
        for spg_type in self.spg_types:
            type_en, _ = self.spg_type_schema_info_zh[spg_type]
            for v in spg_type.properties.values():
                re_list.append(v.name_zh)
                self.properties_mapper[v.name_zh] = spg_type
            for v in spg_type.relations.values():
                v_zh, _, object_type = self.relation_info_en[type_en][v.name]
                re_list.append(v.name_zh)
                self.relations_mapper[v.name_zh] = [spg_type, object_type]
        self.schema_list = self.multischema_split_by_num(self.split_num, re_list)


class OneKE_KGPrompt(OneKEPrompt):
    template_zh: str = "你是一个图谱实体知识结构化专家。根据输入实体类型(entity type)的schema描述，从文本中抽取出相应的实体实例和其属性信息，不存在的属性不输出, 属性存在多值就返回列表，并输出为可解析的json格式。"
    template_en: str = "You are an expert in structured knowledge systems for graph entities. Based on the schema description of the input entity type, you extract the corresponding entity instances and their attribute information from the text. Attributes that do not exist should not be output. If an attribute has multiple values, a list should be returned. The results should be output in a parsable JSON format."

    def __init__(
        self,
        entity_types: List[SPGTypeName],
        language: str = "zh",
        with_description: bool = False,
        split_num: int = 4,
    ):
        super().__init__(
            types_list=entity_types,
            language=language,
            with_description=with_description,
            split_num=split_num,
        )

    def parse_response(self, response: str) -> List[SPGRecord]:
        if isinstance(response, list) and len(response) > 0:
            response = response[0]
        try:
            re_obj = json.loads(response)
        except json.decoder.JSONDecodeError:
            print("DeepKE_KGPrompt response JSONDecodeError error.")
            return []
        if type(re_obj) != dict:
            print("DeepKE_KGPrompt response type error.")
            return []

        spg_records = []
        for type_zh, type_value in re_obj.items():
            if type_zh not in self.spg_type_schema_info_zh:
                print(f"Unrecognized entity_type: {type_zh}")
                continue
            type_en, _ = self.spg_type_schema_info_zh[type_zh]
            if type_value and isinstance(type_value, dict):
                for name, attrs in type_value.items():
                    spg_record = SPGRecord(type_en).upsert_properties(
                        {"id": name, "name": name}
                    )
                    for attr_zh, attr_value in attrs.items():
                        if isinstance(attr_value, list):
                            attr_value = ",".join(attr_value)
                        if attr_zh in self.property_info_zh[type_zh]:
                            attr_en, _, object_type = self.property_info_zh[type_zh][
                                attr_zh
                            ]
                            spg_record.upsert_property(attr_en, attr_value)
                        elif attr_zh in self.relation_info_zh[type_zh]:
                            attr_en, _, object_type = self.relation_info_zh[type_zh][
                                attr_zh
                            ]
                            spg_record.upsert_relation(attr_en, object_type, attr_value)
                        else:
                            print(f"Unrecognized attribute: {attr_zh}")
                            continue
                        if object_type == "Integer":
                            matches = re.findall(r"\d+", attr_value)
                            if matches:
                                spg_record.upsert_property(attr_en, matches[0])
                        elif object_type == "Float":
                            matches = re.findall(r"\d+(?:\.\d+)?", attr_value)
                            if matches:
                                spg_record.upsert_property(attr_en, matches[0])
                    spg_records.append(spg_record)
        return spg_records

    def _render(self):
        spo_list = []
        for spg_type in self.spg_types:
            attributes = []
            attributes.extend(
                [
                    v.name_zh
                    for k, v in spg_type.properties.items()
                    if k not in ["id", "description", "stdId"]
                ]
            )
            attributes.extend(
                [
                    v.name_zh
                    for k, v in spg_type.relations.items()
                    if v not in attributes and k not in ["isA"]
                ]
            )
            entity_type = spg_type.name_zh
            spo_list.append({"entity_type": entity_type, "attributes": attributes})

        self.schema_list = self.multischema_split_by_num(self.split_num, spo_list)


class OneKE_EEPrompt(OneKEPrompt):
    template_zh: str = "你是专门进行事件提取的专家。请从input中抽取出符合schema定义的事件，不存在的事件返回空列表，不存在的论元返回NAN，如果论元存在多值请返回列表。请按照JSON字符串的格式回答。"
    template_en: str = "You are an expert in event extraction. Please extract events from the input that conform to the schema definition. Return an empty list for events that do not exist, and return NAN for arguments that do not exist. If an argument has multiple values, please return a list. Respond in the format of a JSON string."

    def __init__(
        self,
        event_types: List[SPGTypeName],
        language: str = "zh",
        with_description: bool = False,
        split_num: int = 4,
    ):
        super().__init__(
            types_list=event_types,
            language=language,
            with_description=with_description,
            split_num=split_num,
        )

    def parse_response(self, response: str) -> List[SPGRecord]:
        if isinstance(response, list) and len(response) > 0:
            response = response[0]
        try:
            ee_obj = json.loads(response)
        except json.decoder.JSONDecodeError:
            print("DeepKE_EEPrompt response JSONDecodeError error.")
            return []
        if type(ee_obj) != dict:
            print("DeepKE_EEPrompt response type error.")
            return []

        spg_records = []
        for type_zh, type_values in ee_obj.items():
            if type_zh not in self.spg_type_schema_info_zh:
                print(f"Unrecognized event_type: {type_zh}")
                continue
            type_en, _ = self.spg_type_schema_info_zh[type_zh]
            if type_values and isinstance(type_values, list):
                for type_value in type_values:
                    spg_record = SPGRecord(type_en).upsert_property("name", type_zh)
                    arguments = type_value.get("arguments")
                    if arguments and isinstance(arguments, dict):
                        for attr_zh, attr_value in arguments.items():
                            if isinstance(attr_value, list):
                                attr_value = ",".join(attr_value)
                            if attr_zh in self.property_info_zh[type_zh]:
                                attr_en, _, object_type = self.property_info_zh[
                                    type_zh
                                ][attr_zh]
                                spg_record.upsert_property(attr_en, attr_value)
                            elif attr_zh in self.relation_info_zh[type_zh]:
                                attr_en, _, object_type = self.relation_info_zh[
                                    type_zh
                                ][attr_zh]
                                spg_record.upsert_relation(
                                    attr_en, object_type, attr_value
                                )
                            else:
                                print(f"Unrecognized attribute: {attr_zh}")
                                continue
                            if object_type == "Integer":
                                matches = re.findall(r"\d+", attr_value)
                                if matches:
                                    spg_record.upsert_property(attr_en, matches[0])
                            elif object_type == "Float":
                                matches = re.findall(r"\d+(?:\.\d+)?", attr_value)
                                if matches:
                                    spg_record.upsert_property(attr_en, matches[0])
                    spg_records.append(spg_record)
        return spg_records

    def _render(self):
        event_list = []
        for spg_type in self.spg_types:
            arguments = []
            arguments.extend(
                [
                    v.name_zh
                    for k, v in spg_type.properties.items()
                    if k not in ["id", "name", "description"]
                ]
            )
            arguments.extend(
                [
                    v.name_zh
                    for k, v in spg_type.relations.items()
                    if v.name_zh not in arguments
                ]
            )
            event_type = spg_type.name_zh
            event_list.append(
                {"event_type": event_type, "trigger": True, "arguments": arguments}
            )
        self.schema_list = self.multischema_split_by_num(self.split_num, event_list)



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from typing import List, Dict, Union, Optional

from knext.schema import rest
from knext.schema.model.base import (
    BaseSpgType,
    SpgTypeEnum,
    ROOT_TYPE_UNIQUE_NAME,
    HypernymPredicateEnum,
    ConstraintTypeEnum,
)
from knext.schema.model.property import Property
from knext.schema.model.relation import Relation
from knext.builder.operator.op import LinkOp, FuseOp


class EntityType(BaseSpgType):
    """EntityType Model."""

    spg_type_enum: SpgTypeEnum
    name: str
    name_zh: str
    desc: str
    parent_type_name: str
    properties: Dict[str, Property]
    relations: Dict[str, Relation]
    link_operator: LinkOp
    fuse_operator: FuseOp

    def __init__(
        self,
        name: str,
        name_zh: str = None,
        desc: str = None,
        parent_type_name: str = ROOT_TYPE_UNIQUE_NAME,
        properties: List[Property] = None,
        relations: List[Relation] = None,
        link_operator: LinkOp = None,
        fuse_operator: FuseOp = None,
        **kwargs,
    ):
        super().__init__(
            spg_type_enum=SpgTypeEnum.Entity,
            name=name,
            name_zh=name_zh,
            desc=desc,
            properties=properties,
            relations=relations,
            parent_type_name=parent_type_name,
            link_operator=link_operator,
            fuse_operator=fuse_operator,
            **kwargs,
        )
        if "rest_model" not in kwargs:
            self.link_operator = link_operator
            self.fuse_operator = fuse_operator

    @property
    def link_operator(self):
        """Gets the link_operator of this EntityType.  # noqa: E501


        :return: The link_operator of this EntityType.  # noqa: E501
        :rtype: Operator
        """
        if self._rest_model.advanced_config.link_operator is None:
            return None
        op_name = self._rest_model.advanced_config.link_operator.name
        op_version = self._rest_model.advanced_config.link_operator.version
        return self.link_operator.to_rest(op_name, op_version)

    @link_operator.setter
    def link_operator(self, link_operator: LinkOp):
        """Sets the link_operator of this EntityType.


        :param link_operator: The link_operator of this EntityType.  # noqa: E501
        :type: Operator
        """

        if link_operator is None:
            self._rest_model.advanced_config.link_operator = None
            return
        if self._rest_model.advanced_config.link_operator is None:
            self._rest_model.advanced_config.link_operator = rest.OperatorKey()
        self._rest_model.advanced_config.link_operator.name = link_operator.name
        self._rest_model.advanced_config.link_operator.version = link_operator._version

    def bind_link_operator(self, operator: LinkOp):
        """Binds a link operator to this EntityType."""

        self.link_operator = operator
        return self

    @property
    def fuse_operator(self):
        """Gets the fuse_operator of this EntityType.  # noqa: E501


        :return: The fuse_operator of this EntityType.  # noqa: E501
        :rtype: Operator
        """
        if self._rest_model.advanced_config.fuse_operator is None:
            return None
        op_name = self._rest_model.advanced_config.fuse_operator.name
        op_version = self._rest_model.advanced_config.fuse_operator.version
        return self.fuse_operator.to_rest(op_name, op_version)

    @fuse_operator.setter
    def fuse_operator(self, fuse_operator: FuseOp):
        """Sets the fuse_operator of this EntityType.


        :param fuse_operator: The fuse_operator of this EntityType.  # noqa: E501
        :type: Operator
        """

        if fuse_operator is None:
            self._rest_model.advanced_config.fuse_operator = None
            return
        if self._rest_model.advanced_config.fuse_operator is None:
            self._rest_model.advanced_config.fuse_operator = rest.OperatorKey()
        self._rest_model.advanced_config.fuse_operator.name = fuse_operator.name
        self._rest_model.advanced_config.fuse_operator.version = fuse_operator._version

    def bind_fuse_operator(self, operator: FuseOp):
        """Binds a fuse operator to this EntityType."""

        self.fuse_operator = operator
        return self


class ConceptType(BaseSpgType):
    """ConceptType Model."""

    spg_type_enum: SpgTypeEnum
    name: str
    hypernym_predicate: HypernymPredicateEnum
    name_zh: str
    desc: str
    parent_type_name: str
    properties: Dict[str, Property]
    relations: Dict[str, Relation]
    link_operator: LinkOp
    fuse_operator: FuseOp
    taxonomic_type_name: str

    def __init__(
        self,
        name: str,
        hypernym_predicate: HypernymPredicateEnum,
        name_zh: str = None,
        desc: str = None,
        parent_type_name: str = ROOT_TYPE_UNIQUE_NAME,
        properties: List[Property] = None,
        relations: List[Relation] = None,
        link_operator: LinkOp = None,
        taxonomic_type_name: str = None,
        **kwargs,
    ):
        super().__init__(
            spg_type_enum=SpgTypeEnum.Concept,
            name=name,
            name_zh=name_zh,
            desc=desc,
            properties=properties,
            relations=relations,
            parent_type_name=parent_type_name,
            **kwargs,
        )
        if "rest_model" not in kwargs:
            self.hypernym_predicate = hypernym_predicate
            self.taxonomic_type_name = taxonomic_type_name
            self.link_operator = link_operator

    @property
    def hypernym_predicate(self) -> Optional[HypernymPredicateEnum]:
        """Gets the hypernym_predicate of this ConceptType.  # noqa: E501


        :return: The hypernym_predicate of this ConceptType.  # noqa: E501
        :rtype: HypernymPredicateEnum
        """
        hypernym_predicate = self._rest_model.concept_layer_config.hypernym_predicate
        return HypernymPredicateEnum(hypernym_predicate) if hypernym_predicate else None

    @hypernym_predicate.setter
    def hypernym_predicate(self, hypernym_predicate: HypernymPredicateEnum):
        """Sets the hypernym_predicate of this ConceptType.


        :param hypernym_predicate: The hypernym_predicate of this ConceptType.  # noqa: E501
        :type: HypernymPredicateEnum
        """

        self._rest_model.concept_layer_config.hypernym_predicate = hypernym_predicate

    @property
    def taxonomic_type_name(self) -> Optional[str]:
        """Gets the taxonomic_type_name of this SpgType.  # noqa: E501


        :return: The taxonomic_type_name of this SpgType.  # noqa: E501
        :rtype: str
        """
        if self._rest_model.concept_taxonomic_config is None:
            return None
        return self._rest_model.concept_taxonomic_config.taxonomic_type_unique_name.name

    @taxonomic_type_name.setter
    def taxonomic_type_name(self, taxonomic_type_name: str):
        """Sets the taxonomic_type_name of this ConceptType.


        :param taxonomic_type_name: The taxonomic_type_name of this ConceptType.  # noqa: E501
        :type: str
        """
        if taxonomic_type_name is None:
            self._rest_model.concept_taxonomic_config = None
            return
        self._rest_model.concept_taxonomic_config.taxonomic_type_unique_name.name = (
            taxonomic_type_name
        )

    @property
    def link_operator(self):
        """Gets the link_operator of this ConceptType.  # noqa: E501


        :return: The link_operator of this ConceptType.  # noqa: E501
        :rtype: Operator
        """
        if self._rest_model.advanced_config.normalized_operator is None:
            return None
        op_name = self._rest_model.advanced_config.normalized_operator.name
        op_version = self._rest_model.advanced_config.normalized_operator.version
        return self.link_operator(op_name, op_version)

    @link_operator.setter
    def link_operator(self, link_operator: LinkOp):
        """Sets the link_operator of this ConceptType.


        :param link_operator: The link_operator of this ConceptType.  # noqa: E501
        :type: Operator
        """

        if link_operator is None:
            self._rest_model.advanced_config.normalized_operator = None
            return
        if self._rest_model.advanced_config.normalized_operator is None:
            self._rest_model.advanced_config.normalized_operator = rest.OperatorKey()
        self._rest_model.advanced_config.normalized_operator.name = link_operator.name
        self._rest_model.advanced_config.normalized_operator.version = (
            link_operator._version
        )

    def bind_link_operator(self, operator: LinkOp):
        """Binds a property normalize operator to this ConceptType."""

        self.link_operator = operator
        return self

    @property
    def fuse_operator(self):
        """Gets the fuse_operator of this EntityType.  # noqa: E501


        :return: The fuse_operator of this EntityType.  # noqa: E501
        :rtype: Operator
        """
        if self._rest_model.advanced_config.fuse_operator is None:
            return None
        op_name = self._rest_model.advanced_config.fuse_operator.name
        op_version = self._rest_model.advanced_config.fuse_operator.version
        return self.fuse_operator.to_rest(op_name, op_version)

    @fuse_operator.setter
    def fuse_operator(self, fuse_operator: FuseOp):
        """Sets the fuse_operator of this EntityType.


        :param fuse_operator: The fuse_operator of this EntityType.  # noqa: E501
        :type: Operator
        """

        if fuse_operator is None:
            self._rest_model.advanced_config.fuse_operator = None
            return
        if self._rest_model.advanced_config.fuse_operator is None:
            self._rest_model.advanced_config.fuse_operator = rest.OperatorKey()
        self._rest_model.advanced_config.fuse_operator.name = fuse_operator.name
        self._rest_model.advanced_config.fuse_operator.version = fuse_operator._version

    def bind_fuse_operator(self, operator: FuseOp):
        """Binds a fuse operator to this EntityType."""

        self.fuse_operator = operator
        return self


class EventType(BaseSpgType):
    """EventType Model."""

    spg_type_enum: SpgTypeEnum
    name: str
    name_zh: str
    desc: str
    parent_type_name: str
    properties: Dict[str, Property]
    relations: Dict[str, Relation]
    link_operator: LinkOp

    def __init__(
        self,
        name: str,
        name_zh: str = None,
        desc: str = None,
        parent_type_name: str = ROOT_TYPE_UNIQUE_NAME,
        properties: List[Property] = None,
        relations: List[Relation] = None,
        link_operator: LinkOp = None,
        **kwargs,
    ):
        super().__init__(
            spg_type_enum=SpgTypeEnum.Event,
            name=name,
            name_zh=name_zh,
            desc=desc,
            properties=properties,
            relations=relations,
            parent_type_name=parent_type_name,
            **kwargs,
        )
        if "rest_model" not in kwargs:
            self.link_operator = link_operator

    @property
    def link_operator(self):
        """Gets the link_operator of this EventType.  # noqa: E501


        :return: The link_operator of this EventType.  # noqa: E501
        :rtype: Operator
        """
        if self._rest_model.advanced_config.link_operator is None:
            return None
        op_name = self._rest_model.advanced_config.normalized_operator.name
        op_version = self._rest_model.advanced_config.normalized_operator.version
        return self.link_operator(op_name, op_version)

    @link_operator.setter
    def link_operator(self, link_operator: LinkOp):
        """Sets the link_operator of this EventType.


        :param link_operator: The link_operator of this EventType.  # noqa: E501
        :type: Operator
        """

        if link_operator is None:
            self._rest_model.advanced_config.link_operator = None
            return
        if self._rest_model.advanced_config.link_operator is None:
            self._rest_model.advanced_config.link_operator = rest.OperatorKey()
        self._rest_model.advanced_config.link_operator.name = link_operator.name
        self._rest_model.advanced_config.link_operator.version = link_operator._version

    def bind_link_operator(self, operator: LinkOp):
        """Binds a link operator to this EventType."""

        self.link_operator = operator
        return self


class BasicType(BaseSpgType):
    """BasicType Model."""

    Text = BaseSpgType(SpgTypeEnum.Basic, "Text")
    Integer = BaseSpgType(SpgTypeEnum.Basic, "Integer")
    Float = BaseSpgType(SpgTypeEnum.Basic, "Float")

    def __init__(self, name: str, **kwargs):
        super().__init__(spg_type_enum=SpgTypeEnum.Basic, name=name, **kwargs)


class StandardType(BaseSpgType):
    """StandardType Model."""

    spg_type_enum: SpgTypeEnum
    name: str
    parent_type_name: str
    spreadable: bool
    constraint: Dict[ConstraintTypeEnum, Union[str, List[str]]]

    def __init__(
        self,
        name: str,
        parent_type_name: str = ROOT_TYPE_UNIQUE_NAME,
        spreadable: bool = False,
        constraint: Dict[ConstraintTypeEnum, Union[str, List[str]]] = None,
        **kwargs,
    ):
        super().__init__(
            spg_type_enum=SpgTypeEnum.Standard,
            name=name,
            parent_type_name=parent_type_name,
            spreadable=spreadable,
            constraint=constraint,
            **kwargs,
        )

    @property
    def spreadable(self) -> bool:
        """Gets the `spreadable` of this StandardType.  # noqa: E501


        :return: The `spreadable` of this StandardType.  # noqa: E501
        :rtype: bool
        """
        return self._rest_model.spreadable

    @spreadable.setter
    def spreadable(self, spreadable: bool):
        """Sets the `spreadable` of this StandardType.


        :param spreadable: The `spreadable` of this StandardType.  # noqa: E501
        :type: bool
        """
        self._rest_model.spreadable = spreadable

    @property
    def constraint(self) -> Dict[ConstraintTypeEnum, Union[str, list]]:
        """Gets the constraint of this StandardType.  # noqa: E501


        :return: The constraint of this StandardType.  # noqa: E501
        :rtype: dict
        """
        if self._rest_model.constraint_items is None:
            return {}
        constraint = {}
        for item in self._rest_model.constraint_items:
            if item.constraint_type_enum == ConstraintTypeEnum.Enum:
                value = item.enum_values
            elif item.constraint_type_enum == ConstraintTypeEnum.Regular:
                value = item.regular_pattern
            else:
                value = None
            constraint[item.constraint_type_enum] = value
        return constraint

    @constraint.setter
    def constraint(self, constraint: Dict[ConstraintTypeEnum, Union[str, list]]):
        """Sets the constraint of this StandardType.


        :param constraint: The constraint of this StandardType.  # noqa: E501
        :type: dict
        """
        if constraint is None:
            return
        self._rest_model.constraint_items = []
        for type, value in constraint.items():
            self.add_constraint(type, value)

    def add_constraint(self, type: ConstraintTypeEnum, value: Union[str, list] = None):
        """Adds a constraint to this StandardType.


        :param type: The type of constraint to add.
        :type type: ConstraintTypeEnum
        :param value: The value(s) of the constraint. Optional.
        :type value: str or list, optional
        """

        if self._rest_model.constraint_items is None:
            self._rest_model.constraint_items = []
        if type == ConstraintTypeEnum.Enum:
            if not isinstance(value, list):
                raise ValueError("Invalid enum format.")
            constraint_item = rest.EnumConstraint(enum_values=value)
        elif type == ConstraintTypeEnum.Regular:
            constraint_item = rest.RegularConstraint(regular_pattern=value)
        else:
            constraint_item = rest.BaseConstraintItem(type)
        self._rest_model.constraint_items.append(constraint_item)
        return self



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from knext.builder.operator.op import LinkOp, ExtractOp, FuseOp, PromptOp, PredictOp
from knext.builder.operator.spg_record import SPGRecord
from knext.builder.operator.builtin.auto_prompt import REPrompt, EEPrompt
from knext.builder.operator.builtin.deepke_prompt import (
    OneKE_NERPrompt,
    OneKE_REPrompt,
    OneKE_SPOPrompt,
    OneKE_KGPrompt,
    OneKE_EEPrompt,
)

__all__ = [
    "ExtractOp",
    "LinkOp",
    "FuseOp",
    "PromptOp",
    "PredictOp",
    "SPGRecord",
    "REPrompt",
    "EEPrompt",
    "OneKE_NERPrompt",
    "OneKE_REPrompt",
    "OneKE_SPOPrompt",
    "OneKE_KGPrompt",
    "OneKE_EEPrompt",
]



#
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import json
import re
from abc import ABC
from typing import List, Dict, Tuple


from knext.schema.client import SchemaClient
from knext.schema.model.base import BaseSpgType
from knext.schema.model.schema_helper import SPGTypeName, PropertyName, RelationName
from knext.builder.operator.op import PromptOp
from knext.builder.operator.spg_record import SPGRecord
import uuid


class AutoPrompt(PromptOp, ABC):
    spg_types: List[BaseSpgType]

    def __init__(self, spg_type_names: List[SPGTypeName]):
        super().__init__()
        self.spg_types = []
        schema_session = SchemaClient().create_session()
        for spg_type_name in spg_type_names:
            spg_type = schema_session.get(spg_type_name=spg_type_name)
            self.spg_types.append(spg_type)

    def _init_render_variables(self):
        self.property_info_zh = {}
        self.property_info_en = {}
        self.relation_info_zh = {}
        self.relation_info_en = {}
        self.spg_type_schema_info_en = {
            "Text": ("文本", None),
            "Integer": ("整型", None),
            "Float": ("浮点型", None),
        }
        self.spg_type_schema_info_zh = {
            "文本": ("Text", None),
            "整型": ("Integer", None),
            "浮点型": ("Float", None),
        }
        for spg_type in self.spg_types:
            self.property_info_zh[spg_type.name_zh] = {}
            self.relation_info_zh[spg_type.name_zh] = {}
            self.property_info_en[spg_type.name] = {}
            self.relation_info_en[spg_type.name] = {}
            for _rel in spg_type.relations.values():
                if _rel.is_dynamic:
                    continue
                self.relation_info_zh[spg_type.name_zh][_rel.name_zh] = (
                    _rel.name,
                    _rel.desc,
                    _rel.object_type_name,
                )
                self.relation_info_en[spg_type.name][_rel.name] = (
                    _rel.name_zh,
                    _rel.desc,
                    _rel.object_type_name,
                )
            for _prop in spg_type.properties.values():
                self.property_info_zh[spg_type.name_zh][_prop.name_zh] = (
                    _prop.name,
                    _prop.desc,
                    _prop.object_type_name,
                )
                self.property_info_en[spg_type.name][_prop.name] = (
                    _prop.name_zh,
                    _prop.desc,
                    _prop.object_type_name,
                )
        for _type in self.spg_types:
            self.spg_type_schema_info_zh[_type.name_zh] = (_type.name, _type.desc)
            self.spg_type_schema_info_en[_type.name] = (_type.name_zh, _type.desc)


class REPrompt(AutoPrompt):
    template: str = """
已知SPO关系包括:${schema}
从下列句子中提取定义的这些关系。最终抽取结果以json格式输出，且predicate必须在[${predicate}]内。
input:${input}
输出格式为:{"spo":[{"subject":,"predicate":,"object":},]}
"output":
    """

    def __init__(
        self,
        spg_type_name: SPGTypeName,
        property_names: List[PropertyName] = None,
        relation_names: List[Tuple[RelationName, SPGTypeName]] = None,
        custom_prompt: str = None,
    ):
        super().__init__([spg_type_name])

        self.spg_type_name = spg_type_name
        if custom_prompt:
            self.template = custom_prompt
        if not property_names:
            property_names = []
        if not relation_names:
            relation_names = []

        self.property_names = property_names
        self.relation_names = relation_names

        self._init_render_variables()
        self._render()

    def build_prompt(self, variables: Dict[str, str]) -> str:
        return self.template.replace("${input}", variables.get("input"))

    def parse_response(self, response: str) -> List[SPGRecord]:
        if isinstance(response, list) and len(response) > 0:
            response = response[0]
        re_obj = json.loads(response)
        if "spo" not in re_obj.keys():
            raise ValueError("SPO format error.")
        subject_records = {}
        result = []
        for spo_item in re_obj.get("spo", []):
            if any(k not in spo_item for k in ["subject", "predicate", "object"]):
                continue
            s, p_zh, o = spo_item["subject"], spo_item["predicate"], spo_item["object"]
            if s not in subject_records:
                subject_records[s] = (
                    SPGRecord(spg_type_name=self.spg_type_name)
                    .upsert_property("id", s)
                    .upsert_property("name", s)
                )
            if p_zh in self.property_info_zh:
                p, _, o_type = self.property_info_zh[p_zh]
                o_list = re.split("[,，、;；]", o)
                result.extend(
                    [
                        SPGRecord(o_type)
                        .upsert_property("id", _o)
                        .upsert_property("name", _o)
                        for _o in o_list
                    ]
                )
                o = subject_records[s].get_property(p)
                o = ",".join([o] + o_list) if o else ",".join(o_list)
                subject_records[s].upsert_property(p, o)
            elif p_zh in self.relation_info_zh:
                p, _, o_type = self.relation_info_zh[p_zh]
                if "." not in o_type or "STD." in o_type:
                    continue
                o_list = re.split("[,，、;；]", o)
                result.extend(
                    [
                        SPGRecord(o_type)
                        .upsert_property("id", _o)
                        .upsert_property("name", _o)
                        for _o in o_list
                    ]
                )
                o = subject_records[s].get_relation(p, o_type, "")
                o = ",".join([o] + o_list) if o else ",".join(o_list)
                subject_records[s].upsert_relation(p, o_type, o)
            else:
                continue

        for subject_record in subject_records.values():
            result.append(subject_record)
        return result

    def _render(self):
        spo_infos = []
        predicates = []
        duplicate_types = set()
        duplicate_predicates = set()
        for _prop in self.property_names:
            s_name_zh, s_desc = self.spg_type_schema_info_en.get(self.spg_type_name)
            s_desc = (
                (s_desc or s_name_zh)
                if self.spg_type_name not in duplicate_types
                else None
            )
            s_info = (s_name_zh or "") + (f"({s_desc})" if s_desc else "")
            p_name_zh, p_desc, o_type = self.property_info_en.get(_prop)
            p_desc = (
                (p_desc or p_name_zh) if _prop not in duplicate_predicates else None
            )
            p_info = (p_name_zh or "") + (f"({p_desc})" if p_desc else "")
            o_name_zh, o_desc = self.spg_type_schema_info_en.get(o_type)
            o_desc = (o_desc or o_name_zh) if o_type not in duplicate_types else None
            o_info = (o_name_zh or "") + (f"({o_desc})" if o_desc else "")
            spo_infos.append(f"{s_info}-{p_info}-{o_info}")
            duplicate_predicates.add(_prop)
            duplicate_types.update([self.spg_type_name, o_type])
            predicates.append(p_name_zh)
        for _rel, o_type in self.relation_names:
            s_name_zh, s_desc = self.spg_type_schema_info_en.get(self.spg_type_name)
            s_desc = (
                (s_desc or s_name_zh)
                if self.spg_type_name not in duplicate_types
                else None
            )
            s_info = (s_name_zh or "") + (f"({s_desc})" if s_desc else "")
            p_name_zh, p_desc, _ = self.relation_info_en.get(_rel)
            p_desc = (p_desc or p_name_zh) if _rel not in duplicate_predicates else None
            p_info = (p_name_zh or "") + (f"({p_desc})" if p_desc else "")
            o_name_zh, o_desc = self.spg_type_schema_info_en.get(o_type)
            o_desc = (o_desc or o_name_zh) if o_type not in duplicate_types else None
            o_info = (o_name_zh or "") + (f"({o_desc})" if o_desc else "")
            spo_infos.append(f"{s_info}-{p_info}-{o_info}")
            duplicate_predicates.add(_rel)
            duplicate_types.update([self.spg_type_name, o_type])
            predicates.append(p_name_zh)
        schema_text = "\n[" + ",\n".join(spo_infos) + "]"
        predicate_text = ",".join(predicates)
        self.template = self.template.replace("${schema}", schema_text).replace(
            "${predicate}", predicate_text
        )


class EEPrompt(AutoPrompt):
    template: str = """
{
'input': '${input}',
'instruction': '你是专门进行事件提取的专家。请从input中抽取出符合schema定义的事件，不存在的事件返回空列表，不存在的论元返回空字符串。
请按照JSON字符串的格式回答。输出格式为:{\"event\":[{\"event_type\":,\"arguments\":{\"\":,},}]}',
'schema': '${schema}',
'schema description': '${description}'
}
    """

    def __init__(
        self,
        event_type_name: SPGTypeName,
        property_names: List[PropertyName] = None,
        relation_names: List[Tuple[RelationName, SPGTypeName]] = None,
        custom_prompt: str = None,
    ):
        super().__init__([event_type_name])

        self.spg_type_name = event_type_name
        if custom_prompt:
            self.template = custom_prompt
        if not property_names:
            property_names = []
        if not relation_names:
            relation_names = []

        self.property_names = property_names
        self.relation_names = relation_names

        self._init_render_variables()
        self._render()

    def build_prompt(self, variables: Dict[str, str]) -> str:
        return self.template.replace("${input}", variables.get("input"))

    def parse_response(self, response: str) -> List[SPGRecord]:
        if isinstance(response, list) and len(response) > 0:
            response = response[0]
        re_obj = json.loads(response)
        if "event" not in re_obj.keys():
            raise ValueError("EEPrompt response format error.")
        subject_records = []
        object_records = []
        for event_item in re_obj.get("event", []):
            if any(k not in event_item for k in ["event_type", "arguments"]):
                continue
            event_type, arguments = event_item["event_type"], event_item["arguments"]

            uuid_4 = uuid.uuid4()
            subject_record = SPGRecord(
                spg_type_name=self.spg_type_name
            ).upsert_property("id", uuid_4.hex)
            for p_zh, o in arguments.items():
                if p_zh in self.property_info_zh:
                    p, _, o_type = self.property_info_zh[p_zh]
                    if not o:
                        continue
                    subject_record.upsert_property(p, o)
                    if "." in o_type and "STD." not in o_type:
                        object_records.append(
                            SPGRecord(o_type)
                            .upsert_property("id", o)
                            .upsert_property("name", o)
                        )
                elif p_zh in self.relation_info_zh:
                    p, _, o_type = self.relation_info_zh[p_zh]
                    if not o:
                        continue
                    subject_record.upsert_relation(p, o_type, o)
                    if "." in o_type or "STD." not in o_type:
                        object_records.append(
                            SPGRecord(o_type)
                            .upsert_property("id", o)
                            .upsert_property("name", o)
                        )
                else:
                    continue
            subject_records.append(subject_record)

        return subject_records + object_records

    def _render(self):
        schema = {}
        description = {}
        arguments = ["名称"]
        for _prop in self.property_names:
            p_name_zh, p_desc, o_type = self.property_info_en.get(_prop)
            o_name_zh, o_desc = self.spg_type_schema_info_en.get(o_type)
            p_desc = (
                f"类型是{o_name_zh}"
                + ("，" + o_desc if o_desc else "")
                + f"。{p_desc or p_name_zh}"
            )
            arguments.append(p_name_zh)
            description[p_name_zh] = p_desc or p_name_zh
        for _rel, o_type in self.relation_names:
            p_name_zh, p_desc, _ = self.relation_info_en.get(_rel)
            o_name_zh, o_desc = self.spg_type_schema_info_en.get(o_type)
            name_zh = p_name_zh + "#" + o_name_zh
            desc = (
                f"类型是{o_name_zh}"
                + ("，" + o_desc if o_desc else "")
                + f"。{p_desc or p_name_zh}"
            )
            arguments.append(p_name_zh)
            description[name_zh] = desc

        s_name_zh, s_desc = self.spg_type_schema_info_en.get(self.spg_type_name)
        schema["arguments"] = arguments
        schema["event_type"] = s_name_zh
        description[s_name_zh] = s_desc or s_name_zh
        self.template = self.template.replace(
            "${schema}", json.dumps(schema, ensure_ascii=False)
        ).replace("${description}", json.dumps(description, ensure_ascii=False))



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os
import sys
from configparser import ConfigParser
from pathlib import Path
from typing import Union, Optional


GLOBAL_CONFIG = ["host_addr", "graph_store_url", "search_engine_url"]
LOCAL_CONFIG = [
    "project_name",
    "project_id",
    "namespace",
    "description",
    "project_dir",
    "schema_dir",
    "schema_file",
    "builder_dir",
    "builder_record_dir",
    "builder_operator_dir",
    "builder_job_dir",
    "builder_model_dir",
    "reasoner_dir",
    "reasoner_result_dir",
]
CFG_PREFIX = "KNEXT_"
LOCAL_SCHEMA_URL = "http://localhost:8887"


def init_env():
    """Initialize environment to use command-line tool from inside a project
    dir. This sets the Scrapy settings module and modifies the Python path to
    be able to locate the project module.
    """
    project_cfg, root_path = get_config()

    if project_cfg.has_section("global"):
        for cfg in GLOBAL_CONFIG:
            if project_cfg.has_option("global", cfg):
                os.environ[CFG_PREFIX + cfg.upper()] = project_cfg.get("global", cfg)

    if project_cfg.has_section("local"):
        for cfg in LOCAL_CONFIG:
            if project_cfg.has_option("local", cfg):
                os.environ[CFG_PREFIX + cfg.upper()] = project_cfg.get("local", cfg)

        os.environ[CFG_PREFIX + "ROOT_PATH"] = str(root_path.resolve())

    load_operator()
    load_builder_job()


def get_config():
    """
    Get knext config file as a ConfigParser.
    """
    global_cfg, _, local_cfg, local_cfg_path = get_cfg_files()

    for section in global_cfg.sections():
        local_cfg.add_section(section)
        for key, value in global_cfg.items(section):
            local_cfg.set(section, key, value)

    return local_cfg, Path(local_cfg_path).parent


def _closest_cfg(
    path: Union[str, os.PathLike] = ".",
    prev_path: Optional[Union[str, os.PathLike]] = None,
) -> str:
    """
    Return the path to the closest .knext.cfg file by traversing the current
    directory and its parents
    """
    if prev_path is not None and str(path) == str(prev_path):
        return ""
    path = Path(path).resolve()
    cfg_file = path / ".knext.cfg"
    if cfg_file.exists():
        return str(cfg_file)
    return _closest_cfg(path.parent, path)


def get_cfg_files():
    """
    Get global and local knext config files and paths.
    """
    global_cfg_path = (
        Path(os.environ.get("XDG_CONFIG_HOME") or "~/.config").expanduser()
    ) / ".knext.cfg"
    if not global_cfg_path.parent.exists():
        Path.mkdir(global_cfg_path.parent)
    global_cfg = ConfigParser()
    global_cfg.read(global_cfg_path)
    local_cfg_path = _closest_cfg()
    local_cfg = ConfigParser()
    local_cfg.read(local_cfg_path)

    if local_cfg_path:
        projdir = str(Path(local_cfg_path).parent)
        if projdir not in sys.path:
            sys.path.append(projdir)

    return global_cfg, global_cfg_path, local_cfg, local_cfg_path


def load_operator():
    """
    Load all operators in [builder_operator_dir].
    """
    from knext.builder.operator.base import BaseOp
    from knext.builder.operator import builtin

    if not BaseOp._has_registered and (
        "KNEXT_ROOT_PATH" in os.environ and "KNEXT_BUILDER_OPERATOR_DIR" in os.environ
    ):

        from knext.common.class_register import register_from_package

        builder_operator_path = os.path.join(
            os.environ["KNEXT_ROOT_PATH"], os.environ["KNEXT_BUILDER_OPERATOR_DIR"]
        )

        register_from_package(builder_operator_path, BaseOp)
        sys.path.append(builtin.__path__[0])


def load_builder_job():
    """
    Load all builder jobs in [builder_job_dir].
    """
    from knext.builder.model.builder_job import BuilderJob

    if not BuilderJob._has_registered and (
        "KNEXT_ROOT_PATH" in os.environ and "KNEXT_BUILDER_JOB_DIR" in os.environ
    ):

        from knext.common.class_register import register_from_package

        builder_operator_path = os.path.join(
            os.environ["KNEXT_ROOT_PATH"], os.environ["KNEXT_BUILDER_JOB_DIR"]
        )

        register_from_package(builder_operator_path, BuilderJob)



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from typing import TypeVar, Sequence, Type

from pydantic import BaseConfig, BaseModel

Other = TypeVar("Other")

Input = TypeVar("Input", contravariant=True)
Output = TypeVar("Output", covariant=True)


class Runnable(BaseModel):
    """
    Abstract base class that can be invoked synchronously.
    """

    _last: bool = False

    @property
    def input_types(self) -> Type[Input]:
        """The type of input this Runnable object accepts specified as a type annotation."""
        return

    @property
    def output_types(self) -> Type[Output]:
        """The type of output this Runnable object produces specified as a type annotation."""
        return

    def invoke(self, input: Input) -> Sequence[Output]:
        """Transform an input into an output sequence synchronously."""
        raise NotImplementedError(
            f"`invoke` is not currently supported for {self.__class__.__name__}."
        )

    def __rshift__(self, other: Other):
        raise NotImplementedError("To be implemented in subclass")

    class Config(BaseConfig):
        arbitrary_types_allowed = True



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import re
import string
import sys
from configparser import ConfigParser
from pathlib import Path
from typing import Optional

import click
from tabulate import tabulate

from knext.common import project as rest
from knext.common.template import copytree, render_template

TEMPLATES_TO_RENDER = (
    (".knext.cfg.tmpl",),
    ("README.md.tmpl",),
    ("schema", "${project}.schema.tmpl"),
    ("reasoner", "company.dsl.tmpl"),
    ("builder", "operator", "address_extract_op.py.tmpl"),
    ("builder", "job", "data", "company.csv"),
    ("builder", "job", "company.py.tmpl"),
)


def list_project():
    """
    List all project information.
    """
    client = rest.ProjectApi()
    projects = client.project_get()

    table_data = []
    for project in projects:
        item = project.to_dict()
        table_data.append(
            [item["id"], item["name"], item["namespace"], item["description"]]
        )

    table_headers = ["ID", "Name", "Namespace", "Description"]

    table = tabulate(table_data, table_headers, tablefmt="github")
    click.echo(table)


def _recover_project(prj_path: str):
    """
    Recover project by a project dir path.
    """
    if not Path(prj_path).exists():
        click.secho(f"ERROR: No such directory: {prj_path}", fg="bright_red")
        sys.exit()

    if prj_path not in sys.path:
        sys.path.append(prj_path)
    prj = Path(prj_path).resolve()

    cfg_file = prj / ".knext.cfg"
    cfg = ConfigParser()
    cfg.read(cfg_file)
    project_name = cfg.get("local", "project_name")
    namespace = cfg.get("local", "namespace")
    desc = cfg.get("local", "description")

    client = rest.ProjectApi()
    project_create_request = rest.ProjectCreateRequest(
        name=project_name, desc=desc, namespace=namespace
    )
    project = client.project_create_post(project_create_request=project_create_request)

    cfg.set("local", "project_id", str(project.id))
    with open(cfg_file, "w") as config_file:
        cfg.write(config_file)

    click.secho(
        f"Project [{project_name}] with namespace [{namespace}] was successfully created from [{prj_path}].",
        fg="bright_green",
    )


@click.option("--name", help="Name of project.")
@click.option("--namespace", help="Prefix of project schema.")
@click.option("--desc", help="Description of project.")
@click.option(
    "--prj_path",
    help="If set, project will be created according to config file of this path.",
)
def create_project(
    name: str, namespace: str, desc: Optional[str], prj_path: Optional[str]
):
    """
    Create new project with a demo case.
    """

    if prj_path:
        _recover_project(prj_path)
        sys.exit()

    if not name:
        click.secho("ERROR: Option [--name] is required.")
        sys.exit()
    if not namespace:
        click.secho("ERROR: Option [--namespace] is required.")
        sys.exit()

    if not re.match(r"^[A-Z][A-Za-z0-9]{0,15}$", namespace):
        raise click.BadParameter(
            f"Invalid namespace: {namespace}."
            f" Must start with an uppercase letter, only contain letters and numbers, and have a maximum length of 16."
        )

    project_dir = Path(namespace.lower())
    if project_dir.exists():
        raise click.ClickException(
            f"Project directory [{namespace.lower()}] already exists."
        )
    client = rest.ProjectApi()
    project_create_request = rest.ProjectCreateRequest(
        name=name, desc=desc, namespace=namespace
    )
    project = client.project_create_post(project_create_request=project_create_request)

    copytree(Path("project"), project_dir.resolve(), namespace.lower())
    for paths in TEMPLATES_TO_RENDER:
        tplfile = Path(
            *(string.Template(s).substitute(project=namespace.lower()) for s in paths),
        )
        render_template(
            project_dir,
            tplfile,
            project_name=name,
            namespace=namespace,
            project_dir=namespace.lower(),
            project_id=project.id,
            description=desc,
            helper=f"{namespace.lower()}_schema_helper",
        )

    click.secho(
        f"Project [{name}] with namespace [{namespace}] was successfully created in {project_dir.resolve()} \n"
        + "You can checkout your project with: \n"
        + f"  cd {project_dir}",
        fg="bright_green",
    )



# coding: utf-8
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.


"""
    knext

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import re  # noqa: F401

# python 2 and python 3 compatibility library
import six

from knext.common.rest.api_client import ApiClient
from knext.common.rest.exceptions import ApiTypeError, ApiValueError  # noqa: F401


class SchemaApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

    def schema_alter_schema_post(self, **kwargs):  # noqa: E501
        """alter_schema  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.schema_alter_schema_post(async_req=True)
        >>> result = thread.get()

        :param async_req bool: execute request asynchronously
        :param SchemaAlterRequest schema_alter_request:
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: object
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs["_return_http_data_only"] = True
        return self.schema_alter_schema_post_with_http_info(**kwargs)  # noqa: E501

    def schema_alter_schema_post_with_http_info(self, **kwargs):  # noqa: E501
        """alter_schema  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.schema_alter_schema_post_with_http_info(async_req=True)
        >>> result = thread.get()

        :param async_req bool: execute request asynchronously
        :param SchemaAlterRequest schema_alter_request:
        :param _return_http_data_only: response data without head status code
                                       and headers
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: tuple(object, status_code(int), headers(HTTPHeaderDict))
                 If the method is called asynchronously,
                 returns the request thread.
        """

        local_var_params = locals()

        all_params = ["schema_alter_request"]
        all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
            ]
        )

        for key, val in six.iteritems(local_var_params["kwargs"]):
            if key not in all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method schema_alter_schema_post" % key
                )
            local_var_params[key] = val
        del local_var_params["kwargs"]

        collection_formats = {}

        path_params = {}

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if "schema_alter_request" in local_var_params:
            body_params = local_var_params["schema_alter_request"]
        # HTTP header `Accept`
        header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # HTTP header `Content-Type`
        header_params[
            "Content-Type"
        ] = self.api_client.select_header_content_type(  # noqa: E501
            ["application/json"]
        )  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            "/schema/alterSchema",
            "POST",
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type="object",  # noqa: E501
            auth_settings=auth_settings,
            async_req=local_var_params.get("async_req"),
            _return_http_data_only=local_var_params.get(
                "_return_http_data_only"
            ),  # noqa: E501
            _preload_content=local_var_params.get("_preload_content", True),
            _request_timeout=local_var_params.get("_request_timeout"),
            collection_formats=collection_formats,
        )

    def schema_query_project_schema_get(self, project_id, **kwargs):  # noqa: E501
        """query_project_schema  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.schema_query_project_schema_get(project_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool: execute request asynchronously
        :param int project_id: (required)
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: ProjectSchema
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs["_return_http_data_only"] = True
        return self.schema_query_project_schema_get_with_http_info(
            project_id, **kwargs
        )  # noqa: E501

    def schema_query_project_schema_get_with_http_info(
        self, project_id, **kwargs
    ):  # noqa: E501
        """query_project_schema  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.schema_query_project_schema_get_with_http_info(project_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool: execute request asynchronously
        :param int project_id: (required)
        :param _return_http_data_only: response data without head status code
                                       and headers
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: tuple(ProjectSchema, status_code(int), headers(HTTPHeaderDict))
                 If the method is called asynchronously,
                 returns the request thread.
        """

        local_var_params = locals()

        all_params = ["project_id"]
        all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
            ]
        )

        for key, val in six.iteritems(local_var_params["kwargs"]):
            if key not in all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method schema_query_project_schema_get" % key
                )
            local_var_params[key] = val
        del local_var_params["kwargs"]
        # verify the required parameter 'project_id' is set
        if self.api_client.client_side_validation and (
            "project_id" not in local_var_params
            or local_var_params["project_id"] is None  # noqa: E501
        ):  # noqa: E501
            raise ApiValueError(
                "Missing the required parameter `project_id` when calling `schema_query_project_schema_get`"
            )  # noqa: E501

        collection_formats = {}

        path_params = {}

        query_params = []
        if (
            "project_id" in local_var_params
            and local_var_params["project_id"] is not None
        ):  # noqa: E501
            query_params.append(
                ("projectId", local_var_params["project_id"])
            )  # noqa: E501

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            "/schema/queryProjectSchema",
            "GET",
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type="ProjectSchema",  # noqa: E501
            auth_settings=auth_settings,
            async_req=local_var_params.get("async_req"),
            _return_http_data_only=local_var_params.get(
                "_return_http_data_only"
            ),  # noqa: E501
            _preload_content=local_var_params.get("_preload_content", True),
            _request_timeout=local_var_params.get("_request_timeout"),
            collection_formats=collection_formats,
        )

    def schema_query_relation_get(
        self, s_name, relation, o_name, **kwargs
    ):  # noqa: E501
        """query_relation  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.schema_query_relation_get(s_name, relation, o_name, async_req=True)
        >>> result = thread.get()

        :param async_req bool: execute request asynchronously
        :param str s_name: (required)
        :param str relation: (required)
        :param str o_name: (required)
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Relation
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs["_return_http_data_only"] = True
        return self.schema_query_relation_get_with_http_info(
            s_name, relation, o_name, **kwargs
        )  # noqa: E501

    def schema_query_relation_get_with_http_info(
        self, s_name, relation, o_name, **kwargs
    ):  # noqa: E501
        """query_relation  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.schema_query_relation_get_with_http_info(s_name, relation, o_name, async_req=True)
        >>> result = thread.get()

        :param async_req bool: execute request asynchronously
        :param str s_name: (required)
        :param str relation: (required)
        :param str o_name: (required)
        :param _return_http_data_only: response data without head status code
                                       and headers
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: tuple(Relation, status_code(int), headers(HTTPHeaderDict))
                 If the method is called asynchronously,
                 returns the request thread.
        """

        local_var_params = locals()

        all_params = ["s_name", "relation", "o_name"]
        all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
            ]
        )

        for key, val in six.iteritems(local_var_params["kwargs"]):
            if key not in all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method schema_query_relation_get" % key
                )
            local_var_params[key] = val
        del local_var_params["kwargs"]
        # verify the required parameter 's_name' is set
        if self.api_client.client_side_validation and (
            "s_name" not in local_var_params
            or local_var_params["s_name"] is None  # noqa: E501
        ):  # noqa: E501
            raise ApiValueError(
                "Missing the required parameter `s_name` when calling `schema_query_relation_get`"
            )  # noqa: E501
        # verify the required parameter 'relation' is set
        if self.api_client.client_side_validation and (
            "relation" not in local_var_params
            or local_var_params["relation"] is None  # noqa: E501
        ):  # noqa: E501
            raise ApiValueError(
                "Missing the required parameter `relation` when calling `schema_query_relation_get`"
            )  # noqa: E501
        # verify the required parameter 'o_name' is set
        if self.api_client.client_side_validation and (
            "o_name" not in local_var_params
            or local_var_params["o_name"] is None  # noqa: E501
        ):  # noqa: E501
            raise ApiValueError(
                "Missing the required parameter `o_name` when calling `schema_query_relation_get`"
            )  # noqa: E501

        collection_formats = {}

        path_params = {}

        query_params = []
        if (
            "s_name" in local_var_params and local_var_params["s_name"] is not None
        ):  # noqa: E501
            query_params.append(("sName", local_var_params["s_name"]))  # noqa: E501
        if (
            "relation" in local_var_params and local_var_params["relation"] is not None
        ):  # noqa: E501
            query_params.append(
                ("relation", local_var_params["relation"])
            )  # noqa: E501
        if (
            "o_name" in local_var_params and local_var_params["o_name"] is not None
        ):  # noqa: E501
            query_params.append(("oName", local_var_params["o_name"]))  # noqa: E501

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            "/schema/queryRelation",
            "GET",
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type="Relation",  # noqa: E501
            auth_settings=auth_settings,
            async_req=local_var_params.get("async_req"),
            _return_http_data_only=local_var_params.get(
                "_return_http_data_only"
            ),  # noqa: E501
            _preload_content=local_var_params.get("_preload_content", True),
            _request_timeout=local_var_params.get("_request_timeout"),
            collection_formats=collection_formats,
        )

    def schema_query_spg_type_get(self, name, **kwargs):  # noqa: E501
        """query_spg_type  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.schema_query_spg_type_get(name, async_req=True)
        >>> result = thread.get()

        :param async_req bool: execute request asynchronously
        :param str name: 实体类型名称 (required)
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: BaseSpgType
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs["_return_http_data_only"] = True
        return self.schema_query_spg_type_get_with_http_info(
            name, **kwargs
        )  # noqa: E501

    def schema_query_spg_type_get_with_http_info(self, name, **kwargs):  # noqa: E501
        """query_spg_type  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.schema_query_spg_type_get_with_http_info(name, async_req=True)
        >>> result = thread.get()

        :param async_req bool: execute request asynchronously
        :param str name: 实体类型名称 (required)
        :param _return_http_data_only: response data without head status code
                                       and headers
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: tuple(BaseSpgType, status_code(int), headers(HTTPHeaderDict))
                 If the method is called asynchronously,
                 returns the request thread.
        """

        local_var_params = locals()

        all_params = ["name"]
        all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
            ]
        )

        for key, val in six.iteritems(local_var_params["kwargs"]):
            if key not in all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method schema_query_spg_type_get" % key
                )
            local_var_params[key] = val
        del local_var_params["kwargs"]
        # verify the required parameter 'name' is set
        if self.api_client.client_side_validation and (
            "name" not in local_var_params
            or local_var_params["name"] is None  # noqa: E501
        ):  # noqa: E501
            raise ApiValueError(
                "Missing the required parameter `name` when calling `schema_query_spg_type_get`"
            )  # noqa: E501

        collection_formats = {}

        path_params = {}

        query_params = []
        if (
            "name" in local_var_params and local_var_params["name"] is not None
        ):  # noqa: E501
            query_params.append(("name", local_var_params["name"]))  # noqa: E501

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            "/schema/querySpgType",
            "GET",
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type="BaseSpgType",  # noqa: E501
            auth_settings=auth_settings,
            async_req=local_var_params.get("async_req"),
            _return_http_data_only=local_var_params.get(
                "_return_http_data_only"
            ),  # noqa: E501
            _preload_content=local_var_params.get("_preload_content", True),
            _request_timeout=local_var_params.get("_request_timeout"),
            collection_formats=collection_formats,
        )



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import pprint
from typing import TypeVar, Generic, List, Dict, Any

T = TypeVar("T")


class InvokeResult(Generic[T]):
    """Data structure in operator, used to represent the results of operator processing.
    Args:
        data: The data after operator process.
        traces: The user defined logs in operator.
        errors: The exceptions in operator.
    """

    def __init__(self, data: T, traces: List[str] = None, errors: List[str] = None):
        self.data = data
        self.traces = traces
        self.errors = errors

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def to_dict(self):
        """Returns the model properties as a dict"""
        return {
            "data": self.data
            if isinstance(self.data, str)
            else [data.to_dict() for data in self.data if data]
            if self.data
            else [],
            "traces": self.traces,
            "errors": self.errors,
        }

    @staticmethod
    def from_dict(input: Dict[str, Any]):
        """Returns the model from a dict"""
        return InvokeResult(input.get("data"), input.get("traces"), input.get("errors"))

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import click
from knext.builder.model.builder_job import BuilderJob


@click.argument("job_names", required=True)
def execute_job(job_names):
    job_list = [name.strip() for name in job_names.split(",") if name]

    for job in job_list:
        builder_job = BuilderJob.by_name(job)()
        builder_chain = builder_job.build()
        params = {
            param: getattr(builder_job, param)
            for param in builder_job.__annotations__
            if hasattr(builder_job, param) and not param.startswith("_")
        }
        builder_chain.invoke(builder_chain, **params)



# coding: utf-8
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.


"""
    knext

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from knext.common.rest.configuration import Configuration


class OperatorVersion(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        "overview_id": "int",
        "main_class": "str",
        "file_path": "str",
        "version": "int",
    }

    attribute_map = {
        "overview_id": "overviewId",
        "main_class": "mainClass",
        "file_path": "filePath",
        "version": "version",
    }

    def __init__(
        self,
        overview_id=None,
        main_class=None,
        file_path=None,
        version=None,
        local_vars_configuration=None,
    ):  # noqa: E501
        """OperatorVersion - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._overview_id = None
        self._main_class = None
        self._file_path = None
        self._version = None
        self.discriminator = None

        self.overview_id = overview_id
        self.main_class = main_class
        self.file_path = file_path
        self.version = version

    @property
    def overview_id(self):
        """Gets the overview_id of this OperatorVersion.  # noqa: E501


        :return: The overview_id of this OperatorVersion.  # noqa: E501
        :rtype: int
        """
        return self._overview_id

    @overview_id.setter
    def overview_id(self, overview_id):
        """Sets the overview_id of this OperatorVersion.


        :param overview_id: The overview_id of this OperatorVersion.  # noqa: E501
        :type: int
        """
        if (
            self.local_vars_configuration.client_side_validation and overview_id is None
        ):  # noqa: E501
            raise ValueError(
                "Invalid value for `overview_id`, must not be `None`"
            )  # noqa: E501

        self._overview_id = overview_id

    @property
    def main_class(self):
        """Gets the main_class of this OperatorVersion.  # noqa: E501


        :return: The main_class of this OperatorVersion.  # noqa: E501
        :rtype: str
        """
        return self._main_class

    @main_class.setter
    def main_class(self, main_class):
        """Sets the main_class of this OperatorVersion.


        :param main_class: The main_class of this OperatorVersion.  # noqa: E501
        :type: str
        """
        if (
            self.local_vars_configuration.client_side_validation and main_class is None
        ):  # noqa: E501
            raise ValueError(
                "Invalid value for `main_class`, must not be `None`"
            )  # noqa: E501

        self._main_class = main_class

    @property
    def file_path(self):
        """Gets the file_path of this OperatorVersion.  # noqa: E501


        :return: The file_path of this OperatorVersion.  # noqa: E501
        :rtype: str
        """
        return self._file_path

    @file_path.setter
    def file_path(self, file_path):
        """Sets the file_path of this OperatorVersion.


        :param file_path: The file_path of this OperatorVersion.  # noqa: E501
        :type: str
        """
        if (
            self.local_vars_configuration.client_side_validation and file_path is None
        ):  # noqa: E501
            raise ValueError(
                "Invalid value for `file_path`, must not be `None`"
            )  # noqa: E501

        self._file_path = file_path

    @property
    def version(self):
        """Gets the version of this OperatorVersion.  # noqa: E501


        :return: The version of this OperatorVersion.  # noqa: E501
        :rtype: int
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this OperatorVersion.


        :param version: The version of this OperatorVersion.  # noqa: E501
        :type: int
        """
        if (
            self.local_vars_configuration.client_side_validation and version is None
        ):  # noqa: E501
            raise ValueError(
                "Invalid value for `version`, must not be `None`"
            )  # noqa: E501

        self._version = version

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(
                    map(
                        lambda item: (item[0], item[1].to_dict())
                        if hasattr(item[1], "to_dict")
                        else item,
                        value.items(),
                    )
                )
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OperatorVersion):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OperatorVersion):
            return True

        return self.to_dict() != other.to_dict()



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import os
from json import JSONDecodeError
from typing import Any

import json

import click
from click import Context, Group

from knext.common.rest import ApiException


class _ApiExceptionHandler(Group):
    """Echo exceptions."""

    def invoke(self, ctx: Context) -> Any:
        if os.getenv("KNEXT_DEBUG_MODE", "False") == "True":
            return super().invoke(ctx)
        try:
            return super().invoke(ctx)
        except ApiException as api:
            try:
                body = json.loads(api.body)
            except JSONDecodeError:
                raise api
            click.secho("ERROR: " + body, fg="bright_red")
        except Exception as e:
            click.secho("ERROR: " + e.__str__(), fg="bright_red")



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from __future__ import absolute_import

from knext.schema.rest.models.basic_info import BasicInfo
from knext.schema.rest.models.ontology_id import OntologyId
from knext.schema.rest.models.base_ontology import BaseOntology
from knext.schema.rest.models.user_info import UserInfo
from knext.schema.rest.models.predicate.relation import Relation
from knext.schema.rest.models.predicate.property import Property
from knext.schema.rest.models.predicate.property_ref import PropertyRef
from knext.schema.rest.models.predicate.mounted_concept_config import (
    MountedConceptConfig,
)
from knext.schema.rest.models.predicate.property_advanced_config import (
    PropertyAdvancedConfig,
)
from knext.schema.rest.models.predicate.sub_property import SubProperty
from knext.schema.rest.models.predicate.property_ref_basic_info import (
    PropertyRefBasicInfo,
)
from knext.schema.rest.models.predicate.sub_property_basic_info import (
    SubPropertyBasicInfo,
)
from knext.schema.rest.models.alter.schema_draft import SchemaDraft
from knext.schema.rest.models.alter.schema_alter_request import SchemaAlterRequest
from knext.schema.rest.models.type.base_spg_type import BaseSpgType
from knext.schema.rest.models.type.operator_key import OperatorKey
from knext.schema.rest.models.type.event_type import EventType
from knext.schema.rest.models.type.spg_type_ref_basic_info import SpgTypeRefBasicInfo
from knext.schema.rest.models.type.entity_type import EntityType
from knext.schema.rest.models.type.spg_type_advanced_config import SpgTypeAdvancedConfig
from knext.schema.rest.models.type.concept_type import ConceptType
from knext.schema.rest.models.type.base_advanced_type import BaseAdvancedType
from knext.schema.rest.models.type.concept_layer_config import ConceptLayerConfig
from knext.schema.rest.models.type.multi_version_config import MultiVersionConfig
from knext.schema.rest.models.type.standard_type_basic_info import StandardTypeBasicInfo
from knext.schema.rest.models.type.parent_type_info import ParentTypeInfo
from knext.schema.rest.models.type.project_schema import ProjectSchema
from knext.schema.rest.models.type.concept_taxonomic_config import (
    ConceptTaxonomicConfig,
)
from knext.schema.rest.models.type.standard_type import StandardType
from knext.schema.rest.models.type.spg_type_ref import SpgTypeRef
from knext.schema.rest.models.type.basic_type import BasicType
from knext.schema.rest.models.identifier.spg_type_identifier import SpgTypeIdentifier
from knext.schema.rest.models.identifier.base_spg_identifier import BaseSpgIdentifier
from knext.schema.rest.models.identifier.concept_identifier import ConceptIdentifier
from knext.schema.rest.models.identifier.operator_identifier import OperatorIdentifier
from knext.schema.rest.models.identifier.spg_triple_identifier import (
    SpgTripleIdentifier,
)
from knext.schema.rest.models.identifier.predicate_identifier import PredicateIdentifier
from knext.schema.rest.models.concept.remove_logical_causation_request import (
    RemoveLogicalCausationRequest,
)
from knext.schema.rest.models.concept.define_logical_causation_request import (
    DefineLogicalCausationRequest,
)
from knext.schema.rest.models.concept.remove_dynamic_taxonomy_request import (
    RemoveDynamicTaxonomyRequest,
)
from knext.schema.rest.models.concept.define_dynamic_taxonomy_request import (
    DefineDynamicTaxonomyRequest,
)
from knext.schema.rest.models.semantic.base_semantic import BaseSemantic
from knext.schema.rest.models.semantic.predicate_semantic import PredicateSemantic
from knext.schema.rest.models.semantic.rule_code import RuleCode
from knext.schema.rest.models.semantic.logical_rule import LogicalRule
from knext.schema.rest.models.operator.operator_version_response import (
    OperatorVersionResponse,
)
from knext.schema.rest.models.operator.operator_version_request import (
    OperatorVersionRequest,
)
from knext.schema.rest.models.operator.operator_overview import OperatorOverview
from knext.schema.rest.models.operator.operator_version import OperatorVersion
from knext.schema.rest.models.operator.operator_create_response import (
    OperatorCreateResponse,
)
from knext.schema.rest.models.operator.operator_create_request import (
    OperatorCreateRequest,
)
from knext.schema.rest.models.constraint.constraint import Constraint
from knext.schema.rest.models.constraint.base_constraint_item import BaseConstraintItem
from knext.schema.rest.models.constraint.multi_val_constraint import MultiValConstraint
from knext.schema.rest.models.constraint.regular_constraint import RegularConstraint
from knext.schema.rest.models.constraint.not_null_constraint import NotNullConstraint
from knext.schema.rest.models.constraint.enum_constraint import EnumConstraint



# coding: utf-8
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.


"""
    knext

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from knext.common.rest.configuration import Configuration


class BuilderJobInst(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        "job_inst_id": "str",
        "job_id": "str",
        "project_id": "str",
        "status": "str",
        "start_time": "str",
        "end_time": "str",
        "result": "BaseBuilderResult",
        "progress": "str",
        "log_info": "str",
        "external_job_inst_id": "str",
    }

    attribute_map = {
        "job_inst_id": "jobInstId",
        "job_id": "jobId",
        "project_id": "projectId",
        "status": "status",
        "start_time": "startTime",
        "end_time": "endTime",
        "result": "result",
        "progress": "progress",
        "log_info": "logInfo",
        "external_job_inst_id": "externalJobInstId",
    }

    def __init__(
        self,
        job_inst_id=None,
        job_id=None,
        project_id=None,
        status=None,
        start_time=None,
        end_time=None,
        result=None,
        progress=None,
        log_info=None,
        external_job_inst_id=None,
        local_vars_configuration=None,
    ):  # noqa: E501
        """BuilderJobInst - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._job_inst_id = None
        self._job_id = None
        self._project_id = None
        self._status = None
        self._start_time = None
        self._end_time = None
        self._result = None
        self._progress = None
        self._log_info = None
        self._external_job_inst_id = None
        self.discriminator = None

        if job_inst_id is not None:
            self.job_inst_id = job_inst_id
        if job_id is not None:
            self.job_id = job_id
        if project_id is not None:
            self.project_id = project_id
        if status is not None:
            self.status = status
        if start_time is not None:
            self.start_time = start_time
        if end_time is not None:
            self.end_time = end_time
        if result is not None:
            self.result = result
        if progress is not None:
            self.progress = progress
        if log_info is not None:
            self.log_info = log_info
        if external_job_inst_id is not None:
            self.external_job_inst_id = external_job_inst_id

    @property
    def job_inst_id(self):
        """Gets the job_inst_id of this BuilderJobInst.  # noqa: E501


        :return: The job_inst_id of this BuilderJobInst.  # noqa: E501
        :rtype: str
        """
        return self._job_inst_id

    @job_inst_id.setter
    def job_inst_id(self, job_inst_id):
        """Sets the job_inst_id of this BuilderJobInst.


        :param job_inst_id: The job_inst_id of this BuilderJobInst.  # noqa: E501
        :type: str
        """

        self._job_inst_id = job_inst_id

    @property
    def job_id(self):
        """Gets the job_id of this BuilderJobInst.  # noqa: E501


        :return: The job_id of this BuilderJobInst.  # noqa: E501
        :rtype: str
        """
        return self._job_id

    @job_id.setter
    def job_id(self, job_id):
        """Sets the job_id of this BuilderJobInst.


        :param job_id: The job_id of this BuilderJobInst.  # noqa: E501
        :type: str
        """

        self._job_id = job_id

    @property
    def project_id(self):
        """Gets the project_id of this BuilderJobInst.  # noqa: E501


        :return: The project_id of this BuilderJobInst.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this BuilderJobInst.


        :param project_id: The project_id of this BuilderJobInst.  # noqa: E501
        :type: str
        """

        self._project_id = project_id

    @property
    def status(self):
        """Gets the status of this BuilderJobInst.  # noqa: E501


        :return: The status of this BuilderJobInst.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this BuilderJobInst.


        :param status: The status of this BuilderJobInst.  # noqa: E501
        :type: str
        """
        allowed_values = [
            "INIT",
            "QUEUE",
            "RUNNING",
            "CANCEL",
            "FAILURE",
            "SUCCESS",
        ]  # noqa: E501
        if (
            self.local_vars_configuration.client_side_validation
            and status not in allowed_values
        ):  # noqa: E501
            raise ValueError(
                "Invalid value for `status` ({0}), must be one of {1}".format(  # noqa: E501
                    status, allowed_values
                )
            )

        self._status = status

    @property
    def start_time(self):
        """Gets the start_time of this BuilderJobInst.  # noqa: E501


        :return: The start_time of this BuilderJobInst.  # noqa: E501
        :rtype: str
        """
        return self._start_time

    @start_time.setter
    def start_time(self, start_time):
        """Sets the start_time of this BuilderJobInst.


        :param start_time: The start_time of this BuilderJobInst.  # noqa: E501
        :type: str
        """

        self._start_time = start_time

    @property
    def end_time(self):
        """Gets the end_time of this BuilderJobInst.  # noqa: E501


        :return: The end_time of this BuilderJobInst.  # noqa: E501
        :rtype: str
        """
        return self._end_time

    @end_time.setter
    def end_time(self, end_time):
        """Sets the end_time of this BuilderJobInst.


        :param end_time: The end_time of this BuilderJobInst.  # noqa: E501
        :type: str
        """

        self._end_time = end_time

    @property
    def result(self):
        """Gets the result of this BuilderJobInst.  # noqa: E501


        :return: The result of this BuilderJobInst.  # noqa: E501
        :rtype: BaseBuilderResult
        """
        return self._result

    @result.setter
    def result(self, result):
        """Sets the result of this BuilderJobInst.


        :param result: The result of this BuilderJobInst.  # noqa: E501
        :type: BaseBuilderResult
        """

        self._result = result

    @property
    def progress(self):
        """Gets the progress of this BuilderJobInst.  # noqa: E501


        :return: The progress of this BuilderJobInst.  # noqa: E501
        :rtype: str
        """
        return self._progress

    @progress.setter
    def progress(self, progress):
        """Sets the progress of this BuilderJobInst.


        :param progress: The progress of this BuilderJobInst.  # noqa: E501
        :type: str
        """

        self._progress = progress

    @property
    def log_info(self):
        """Gets the log_info of this BuilderJobInst.  # noqa: E501


        :return: The log_info of this BuilderJobInst.  # noqa: E501
        :rtype: str
        """
        return self._log_info

    @log_info.setter
    def log_info(self, log_info):
        """Sets the log_info of this BuilderJobInst.


        :param log_info: The log_info of this BuilderJobInst.  # noqa: E501
        :type: str
        """

        self._log_info = log_info

    @property
    def external_job_inst_id(self):
        """Gets the external_job_inst_id of this BuilderJobInst.  # noqa: E501


        :return: The external_job_inst_id of this BuilderJobInst.  # noqa: E501
        :rtype: str
        """
        return self._external_job_inst_id

    @external_job_inst_id.setter
    def external_job_inst_id(self, external_job_inst_id):
        """Sets the external_job_inst_id of this BuilderJobInst.


        :param external_job_inst_id: The external_job_inst_id of this BuilderJobInst.  # noqa: E501
        :type: str
        """

        self._external_job_inst_id = external_job_inst_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(
                    map(
                        lambda item: (item[0], item[1].to_dict())
                        if hasattr(item[1], "to_dict")
                        else item,
                        value.items(),
                    )
                )
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, BuilderJobInst):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, BuilderJobInst):
            return True

        return self.to_dict() != other.to_dict()



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import os
from abc import ABC

from knext.common import rest


class Client(ABC):

    _rest_client: rest.BaseApi

    def __init__(self, host_addr: str = None, project_id: int = None):
        self._host_addr = host_addr or os.environ.get("KNEXT_HOST_ADDR")
        self._project_id = project_id or os.environ.get("KNEXT_PROJECT_ID")

    @staticmethod
    def serialize(obj):
        return rest.ApiClient().sanitize_for_serialization(obj)



# coding: utf-8
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.


"""
    knext

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from knext.common.rest.configuration import Configuration


class CsvSourceNodeConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        "type": "str",
        "start_row": "int",
        "url": "str",
        "columns": "list[str]",
    }

    attribute_map = {
        "type": "type",
        "start_row": "startRow",
        "url": "url",
        "columns": "columns",
    }

    def __init__(
        self,
        type="CSV_SOURCE",
        start_row=None,
        url=None,
        columns=None,
        local_vars_configuration=None,
    ):  # noqa: E501
        """CsvSourceNodeConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._type = None
        self._start_row = None
        self._url = None
        self._columns = None
        self.discriminator = type

        self.type = type
        if start_row is not None:
            self.start_row = start_row
        if url is not None:
            self.url = url
        if columns is not None:
            self.columns = columns

    @property
    def type(self):
        """Gets the type of this CsvSourceNodeConfig.  # noqa: E501


        :return: The type of this CsvSourceNodeConfig.  # noqa: E501
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this CsvSourceNodeConfig.


        :param type: The type of this CsvSourceNodeConfig.  # noqa: E501
        :type: str
        """
        if (
            self.local_vars_configuration.client_side_validation and type is None
        ):  # noqa: E501
            raise ValueError(
                "Invalid value for `type`, must not be `None`"
            )  # noqa: E501
        allowed_values = [
            "CSV_SOURCE",
            "SPG_TYPE_MAPPING",
            "RELATION_MAPPING",
            "SUBGRAPH_MAPPING",
            "USER_DEFINED_EXTRACT",
            "LLM_BASED_EXTRACT",
            "GRAPH_SINK",
        ]  # noqa: E501
        if (
            self.local_vars_configuration.client_side_validation
            and type not in allowed_values
        ):  # noqa: E501
            raise ValueError(
                "Invalid value for `type` ({0}), must be one of {1}".format(  # noqa: E501
                    type, allowed_values
                )
            )

        self._type = type

    @property
    def start_row(self):
        """Gets the start_row of this CsvSourceNodeConfig.  # noqa: E501


        :return: The start_row of this CsvSourceNodeConfig.  # noqa: E501
        :rtype: int
        """
        return self._start_row

    @start_row.setter
    def start_row(self, start_row):
        """Sets the start_row of this CsvSourceNodeConfig.


        :param start_row: The start_row of this CsvSourceNodeConfig.  # noqa: E501
        :type: int
        """

        self._start_row = start_row

    @property
    def url(self):
        """Gets the url of this CsvSourceNodeConfig.  # noqa: E501


        :return: The url of this CsvSourceNodeConfig.  # noqa: E501
        :rtype: str
        """
        return self._url

    @url.setter
    def url(self, url):
        """Sets the url of this CsvSourceNodeConfig.


        :param url: The url of this CsvSourceNodeConfig.  # noqa: E501
        :type: str
        """

        self._url = url

    @property
    def columns(self):
        """Gets the columns of this CsvSourceNodeConfig.  # noqa: E501


        :return: The columns of this CsvSourceNodeConfig.  # noqa: E501
        :rtype: list[str]
        """
        return self._columns

    @columns.setter
    def columns(self, columns):
        """Sets the columns of this CsvSourceNodeConfig.


        :param columns: The columns of this CsvSourceNodeConfig.  # noqa: E501
        :type: list[str]
        """

        self._columns = columns

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(
                    map(
                        lambda item: (item[0], item[1].to_dict())
                        if hasattr(item[1], "to_dict")
                        else item,
                        value.items(),
                    )
                )
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CsvSourceNodeConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CsvSourceNodeConfig):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.


# flake8: noqa

"""
    knext

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""

from __future__ import absolute_import

__version__ = "1"

# import apis into sdk package
from knext.common.project.project_api import ProjectApi

# import models into sdk package
from knext.common.project.models.project_create_request import ProjectCreateRequest
from knext.common.project.models.project import Project



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
from abc import ABC
from typing import List, Dict, Any

import knext.common.cache
from knext.schema.model.schema_helper import SPGTypeName, TripletName
from knext.builder.operator.base import BaseOp
from knext.builder.operator.invoke_result import InvokeResult
from knext.builder.operator.spg_record import SPGRecord

cache = knext.common.cache.LinkCache(5000, 60)


class ExtractOp(BaseOp, ABC):
    """Base class for all knowledge extract operators."""

    def __init__(self, params: Dict[str, str] = None):
        super().__init__(params)

    def invoke(self, record: Dict[str, str]) -> List[SPGRecord]:
        raise NotImplementedError(
            f"{self.__class__.__name__} need to implement `invoke` method."
        )

    @staticmethod
    def _pre_process(*inputs):
        return (inputs[0],)

    @staticmethod
    def _post_process(output) -> Dict[str, Any]:
        if isinstance(output, InvokeResult):
            return output.to_dict()
        if isinstance(output, tuple):
            return InvokeResult[List[SPGRecord]](*output[:3]).to_dict()
        else:
            return InvokeResult[List[SPGRecord]](output).to_dict()


class LinkOp(BaseOp, ABC):
    """Base class for all entity link operators."""

    bind_to: SPGTypeName

    bind_schemas: Dict[SPGTypeName, str] = {}

    def __init__(self):
        super().__init__()

    def invoke(self, property: str, subject_record: SPGRecord) -> List[SPGRecord]:
        raise NotImplementedError(
            f"{self.__class__.__name__} need to implement `invoke` method."
        )

    def _handle(self, *inputs) -> Dict[str, Any]:
        _property, subject_record = self._pre_process(*inputs)
        cache_key = str(self.bind_to) + _property
        cache_property = cache.get(cache_key)
        if cache_property:
            output = [
                SPGRecord(spg_type_name=self.bind_to).upsert_property(
                    "id", cache_property
                )
            ]
        else:
            output = self.invoke(_property, subject_record)
        post_output = self._post_process(output)
        return post_output

    @staticmethod
    def _pre_process(*inputs):
        return inputs[0], SPGRecord.from_dict(inputs[1])

    @staticmethod
    def _post_process(output) -> Dict[str, Any]:
        if isinstance(output, InvokeResult):
            return output.to_dict()
        if isinstance(output, tuple):
            return InvokeResult[List[SPGRecord]](*output[:3]).to_dict()
        else:
            return InvokeResult[List[SPGRecord]](output).to_dict()


class FuseOp(BaseOp, ABC):
    """
    Base class for all entity fuse operators.
    """

    """"""
    bind_to: SPGTypeName

    bind_schemas: Dict[SPGTypeName, str] = {}

    def __init__(self):
        super().__init__()

    def link(self, subject_record: SPGRecord) -> SPGRecord:
        raise NotImplementedError(
            f"{self.__class__.__name__} need to implement `link` method."
        )

    def merge(self, subject_record: SPGRecord, linked_record: SPGRecord) -> SPGRecord:
        raise NotImplementedError(
            f"{self.__class__.__name__} need to implement `merge` method."
        )

    def invoke(self, subject_records: List[SPGRecord]) -> List[SPGRecord]:
        records = []
        for record in subject_records:
            cache_key = str(self.bind_to) + record.get_property("id", "")
            linked_record = self.link(record)
            if not linked_record:
                records.append(record)
                continue
            merged_record = self.merge(record, linked_record)
            if merged_record:
                cache.put(cache_key, merged_record.get_property("id", ""))
                records.append(merged_record)
        return records

    @staticmethod
    def _pre_process(*inputs):
        return ([SPGRecord.from_dict(input) for input in inputs[0]],)

    @staticmethod
    def _post_process(output) -> Dict[str, Any]:
        if isinstance(output, InvokeResult):
            return output.to_dict()
        if isinstance(output, tuple):
            return InvokeResult[List[SPGRecord]](*output[:3]).to_dict()
        else:
            return InvokeResult[List[SPGRecord]](output).to_dict()


class PromptOp(BaseOp, ABC):
    """Base class for all prompt operators."""

    template: str

    def __init__(self, **kwargs):
        super().__init__()

    def build_prompt(self, variables: Dict[str, str]) -> str:
        raise NotImplementedError(
            f"{self.__class__.__name__} need to implement `build_prompt` method."
        )

    def parse_response(self, response: str) -> List[SPGRecord]:
        return []

    def invoke(self, *args):
        pass

    def _build_next_variables(
        self, variables: Dict[str, str], response: str
    ) -> List[Dict[str, str]]:
        if isinstance(response, list) and len(response) > 0:
            response = response[0]
        variables.update({f"{self.__class__.__name__}": response})
        return [variables]


class PredictOp(BaseOp, ABC):
    """Base class for all predict operators."""

    bind_to: TripletName

    bind_schemas: Dict[TripletName, str] = {}

    def __init__(self):
        super().__init__()

    def invoke(self, subject_record: SPGRecord) -> List[SPGRecord]:
        raise NotImplementedError(
            f"{self.__class__.__name__} need to implement `invoke` method."
        )

    @staticmethod
    def _pre_process(*inputs):
        return (SPGRecord.from_dict(inputs[0]),)

    @staticmethod
    def _post_process(output) -> Dict[str, Any]:
        if isinstance(output, InvokeResult):
            return output.to_dict()
        if isinstance(output, tuple):
            return InvokeResult[List[SPGRecord]](*output[:3]).to_dict()
        else:
            return InvokeResult[List[SPGRecord]](output).to_dict()



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os
import sys
from pathlib import Path
from typing import Type
import inspect


def _register(root, path, files, class_type):
    relative_path = os.path.relpath(path, root)
    module_prefix = relative_path.replace(".", "").replace("/", ".")
    module_prefix = module_prefix + "." if module_prefix else ""
    for file_name in files:
        if file_name.endswith(".py"):
            module_name = module_prefix + os.path.splitext(file_name)[0]
            import importlib

            module = importlib.import_module(module_name)
            classes = inspect.getmembers(module, inspect.isclass)
            for class_name, class_obj in classes:
                if (
                    issubclass(class_obj, class_type)
                    and inspect.getmodule(class_obj) == module
                ):

                    class_type.register(
                        name=class_name,
                        local_path=os.path.join(path, file_name),
                        module_path=module_name,
                    )(class_obj)


def register_from_package(path: str, class_type: Type) -> None:
    """
    Register all classes under the given package.
    Only registered classes can be recognized by knext.
    """
    if not append_python_path(path):
        return
    for root, dirs, files in os.walk(path):
        _register(path, root, files, class_type)
    class_type._has_registered = True


def append_python_path(path: str) -> bool:
    """
    Append the given path to `sys.path`.
    """
    path = Path(path).resolve()
    path = str(path)
    if path not in sys.path:
        sys.path.append(path)
        return True
    return False



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import os

from knext.common.base.client import Client


class ReasonerClient(Client):
    """SPG Reasoner Client."""

    def __init__(self, host_addr: str = None, project_id: int = None):
        super().__init__(host_addr, project_id)

    def execute(self, dsl_content: str, output_file: str = None):
        """
        Execute a synchronous builder job in local runner.
        """

        import subprocess
        import datetime
        from knext.reasoner import lib
        from knext.common import env

        jar_path = os.path.join(lib.__path__[0], lib.LOCAL_REASONER_JAR)
        default_output_file = (
            f"./{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.csv"
        )

        java_cmd = [
            "java",
            "-jar",
            jar_path,
            "--projectId",
            self._project_id,
            "--query",
            dsl_content,
            "--output",
            output_file or default_output_file,
            "--schemaUrl",
            os.environ.get("KNEXT_HOST_ADDR") or env.LOCAL_SCHEMA_URL,
            "--graphStateClass",
            os.environ.get("KNEXT_GRAPH_STATE_CLASS") or lib.LOCAL_GRAPH_STATE_CLASS,
            "--graphStoreUrl",
            os.environ.get("KNEXT_GRAPH_STORE_URL") or lib.LOCAL_GRAPH_STORE_URL,
        ]

        subprocess.call(java_cmd)



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from abc import ABC
from typing import Union, Tuple, Optional


class SchemaHelper(ABC, str):

    __type_name__: str

    def __init__(self, type_name: str):
        self.__type_name__ = type_name


class SPGTypeHelper(SchemaHelper):
    def __init__(self, type_name: str):
        super().__init__(type_name)


class PropertyHelper(SchemaHelper):
    def __init__(self, type_name: str):
        super().__init__(type_name)


class RelationHelper(SchemaHelper):
    def __init__(self, type_name: str):
        super().__init__(type_name)


class SubPropertyHelper(SchemaHelper):
    def __init__(self, type_name: str):
        super().__init__(type_name)


SPGTypeName = Union[str, SPGTypeHelper]
PropertyName = Union[str, PropertyHelper]
RelationName = Union[str, RelationHelper]
SubPropertyName = Union[str, SubPropertyHelper]
TripletName = Tuple[
    SPGTypeName, Union[PropertyName, RelationName], Optional[SPGTypeName]
]



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

__version__ = "1"

from knext.common.rest.api_client import ApiClient
from knext.common.rest.api_client import BaseApi
from knext.common.rest.configuration import Configuration
from knext.common.rest.exceptions import ApiException
from knext.common.rest.exceptions import ApiKeyError
from knext.common.rest.exceptions import ApiTypeError
from knext.common.rest.exceptions import ApiValueError
from knext.common.rest.exceptions import OpenApiException



# coding: utf-8
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.


"""
    knext

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from knext.common.rest.configuration import Configuration


class BuilderJobSubmitRequest(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        "job_name": "str",
        "project_id": "int",
        "pipeline": "Pipeline",
        "cron": "str",
        "idempotent_id": "str",
        "params": "object",
    }

    attribute_map = {
        "job_name": "jobName",
        "project_id": "projectId",
        "pipeline": "pipeline",
        "cron": "cron",
        "idempotent_id": "idempotentId",
        "params": "params",
    }

    def __init__(
        self,
        job_name=None,
        project_id=None,
        pipeline=None,
        cron=None,
        idempotent_id=None,
        params=None,
        local_vars_configuration=None,
    ):  # noqa: E501
        """BuilderJobSubmitRequest - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._job_name = None
        self._project_id = None
        self._pipeline = None
        self._cron = None
        self._idempotent_id = None
        self._params = None
        self.discriminator = None

        if job_name is not None:
            self.job_name = job_name
        if project_id is not None:
            self.project_id = project_id
        if pipeline is not None:
            self.pipeline = pipeline
        if cron is not None:
            self.cron = cron
        if idempotent_id is not None:
            self.idempotent_id = idempotent_id
        if params is not None:
            self.params = params

    @property
    def job_name(self):
        """Gets the job_name of this BuilderJobSubmitRequest.  # noqa: E501


        :return: The job_name of this BuilderJobSubmitRequest.  # noqa: E501
        :rtype: str
        """
        return self._job_name

    @job_name.setter
    def job_name(self, job_name):
        """Sets the job_name of this BuilderJobSubmitRequest.


        :param job_name: The job_name of this BuilderJobSubmitRequest.  # noqa: E501
        :type: str
        """

        self._job_name = job_name

    @property
    def project_id(self):
        """Gets the project_id of this BuilderJobSubmitRequest.  # noqa: E501


        :return: The project_id of this BuilderJobSubmitRequest.  # noqa: E501
        :rtype: int
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this BuilderJobSubmitRequest.


        :param project_id: The project_id of this BuilderJobSubmitRequest.  # noqa: E501
        :type: int
        """

        self._project_id = project_id

    @property
    def pipeline(self):
        """Gets the pipeline of this BuilderJobSubmitRequest.  # noqa: E501


        :return: The pipeline of this BuilderJobSubmitRequest.  # noqa: E501
        :rtype: Pipeline
        """
        return self._pipeline

    @pipeline.setter
    def pipeline(self, pipeline):
        """Sets the pipeline of this BuilderJobSubmitRequest.


        :param pipeline: The pipeline of this BuilderJobSubmitRequest.  # noqa: E501
        :type: Pipeline
        """

        self._pipeline = pipeline

    @property
    def cron(self):
        """Gets the cron of this BuilderJobSubmitRequest.  # noqa: E501


        :return: The cron of this BuilderJobSubmitRequest.  # noqa: E501
        :rtype: str
        """
        return self._cron

    @cron.setter
    def cron(self, cron):
        """Sets the cron of this BuilderJobSubmitRequest.


        :param cron: The cron of this BuilderJobSubmitRequest.  # noqa: E501
        :type: str
        """
        allowed_values = ["SYNC", "ONCE", "CRON"]  # noqa: E501
        if (
            self.local_vars_configuration.client_side_validation
            and cron not in allowed_values
        ):  # noqa: E501
            raise ValueError(
                "Invalid value for `cron` ({0}), must be one of {1}".format(  # noqa: E501
                    cron, allowed_values
                )
            )

        self._cron = cron

    @property
    def idempotent_id(self):
        """Gets the idempotent_id of this BuilderJobSubmitRequest.  # noqa: E501


        :return: The idempotent_id of this BuilderJobSubmitRequest.  # noqa: E501
        :rtype: str
        """
        return self._idempotent_id

    @idempotent_id.setter
    def idempotent_id(self, idempotent_id):
        """Sets the idempotent_id of this BuilderJobSubmitRequest.


        :param idempotent_id: The idempotent_id of this BuilderJobSubmitRequest.  # noqa: E501
        :type: str
        """
        allowed_values = ["LOCAL", "REMOTE"]  # noqa: E501
        if (
            self.local_vars_configuration.client_side_validation
            and idempotent_id not in allowed_values
        ):  # noqa: E501
            raise ValueError(
                "Invalid value for `idempotent_id` ({0}), must be one of {1}".format(  # noqa: E501
                    idempotent_id, allowed_values
                )
            )

        self._idempotent_id = idempotent_id

    @property
    def params(self):
        """Gets the params of this BuilderJobSubmitRequest.  # noqa: E501


        :return: The params of this BuilderJobSubmitRequest.  # noqa: E501
        :rtype: object
        """
        return self._params

    @params.setter
    def params(self, params):
        """Sets the params of this BuilderJobSubmitRequest.


        :param params: The params of this BuilderJobSubmitRequest.  # noqa: E501
        :type: object
        """

        self._params = params

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(
                    map(
                        lambda item: (item[0], item[1].to_dict())
                        if hasattr(item[1], "to_dict")
                        else item,
                        value.items(),
                    )
                )
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, BuilderJobSubmitRequest):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, BuilderJobSubmitRequest):
            return True

        return self.to_dict() != other.to_dict()



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import json
import os
import sys

from knext.builder import rest
from knext.builder.chain import BuilderChain
from knext.common.base.client import Client
from knext.builder.model.builder_job import BuilderJob, AlterOperationEnum
from knext.common.class_register import register_from_package


class BuilderClient(Client):
    """SPG Builder Client."""

    _rest_client = rest.BuilderApi()

    def __init__(self, host_addr: str = None, project_id: int = None):
        super().__init__(host_addr, project_id)

        if "KNEXT_ROOT_PATH" in os.environ and "KNEXT_BUILDER_JOB_DIR" in os.environ:
            self._builder_job_path = os.path.join(
                os.environ["KNEXT_ROOT_PATH"], os.environ["KNEXT_BUILDER_JOB_DIR"]
            )
            register_from_package(self._builder_job_path, BuilderJob)

    def execute(self, builder_chain: BuilderChain, **kwargs):
        import subprocess
        import datetime
        from knext.builder import lib
        from knext.common import env

        jar_path = os.path.join(lib.__path__[0], lib.LOCAL_BUILDER_JAR)
        dag_config = builder_chain.to_rest()
        pipeline = self.serialize(dag_config)
        log_file_name = f"{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.log"

        java_cmd = [
            "java",
            "-jar",
            jar_path,
            "--projectId",
            self._project_id,
            "--jobName",
            kwargs.get("job_name", "default_job"),
            "--pipeline",
            json.dumps(pipeline),
            "--pythonExec",
            sys.executable,
            "--pythonPaths",
            ";".join(sys.path),
            "--schemaUrl",
            os.environ.get("KNEXT_HOST_ADDR") or env.LOCAL_SCHEMA_URL,
            "--parallelism",
            str(kwargs.get("parallelism", "1")),
            "--alterOperation",
            kwargs.get("alter_operation", AlterOperationEnum.Upsert),
            "--logFile",
            log_file_name,
            "--graphStoreUrl",
            os.environ.get("KNEXT_GRAPH_STORE_URL") or lib.LOCAL_GRAPH_STORE_URL,
            "--searchEngineUrl",
            os.environ.get("KNEXT_SEARCH_ENGINE_URL") or lib.LOCAL_SEARCH_ENGINE_URL,
        ]

        if kwargs.get("lead_to"):
            java_cmd.append("--leadTo")

        if os.getenv("KNEXT_DEBUG_MODE", "False") == "True":
            print_java_cmd = [
                cmd if not cmd.startswith("{") else f"'{cmd}'" for cmd in java_cmd
            ]
            print_java_cmd = [
                cmd if not cmd.count(";") > 0 else f"'{cmd}'" for cmd in print_java_cmd
            ]
            print(json.dumps(" ".join(print_java_cmd))[1:-1].replace("'", '"'))

        subprocess.call(java_cmd)



# coding: utf-8
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.


"""
    knext

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from knext.common.rest.configuration import Configuration


class PropertyAdvancedConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        "multi_version_config": "MultiVersionConfig",
        "mounted_concept_config": "MountedConceptConfig",
        "property_group": "str",
        "constraint": "Constraint",
        "sub_properties": "list[SubProperty]",
        "semantics": "list[PredicateSemantic]",
        "logical_rule": "LogicalRule",
    }

    attribute_map = {
        "multi_version_config": "multiVersionConfig",
        "mounted_concept_config": "MountedConceptConfig",
        "property_group": "propertyGroup",
        "constraint": "constraint",
        "sub_properties": "subProperties",
        "semantics": "semantics",
        "logical_rule": "logicalRule",
    }

    def __init__(
        self,
        multi_version_config=None,
        mounted_concept_config=None,
        property_group=None,
        constraint=None,
        sub_properties=None,
        semantics=None,
        logical_rule=None,
        local_vars_configuration=None,
    ):  # noqa: E501
        """PropertyAdvancedConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._multi_version_config = None
        self._mounted_concept_config = None
        self._property_group = None
        self._constraint = None
        self._sub_properties = None
        self._semantics = None
        self._logical_rule = None
        self.discriminator = None

        if multi_version_config is not None:
            self.multi_version_config = multi_version_config
        if mounted_concept_config is not None:
            self.mounted_concept_config = mounted_concept_config
        if property_group is not None:
            self.property_group = property_group
        if constraint is not None:
            self.constraint = constraint
        if sub_properties is not None:
            self.sub_properties = sub_properties
        if semantics is not None:
            self.semantics = semantics
        if logical_rule is not None:
            self.logical_rule = logical_rule

    @property
    def multi_version_config(self):
        """Gets the multi_version_config of this PropertyAdvancedConfig.  # noqa: E501


        :return: The multi_version_config of this PropertyAdvancedConfig.  # noqa: E501
        :rtype: MultiVersionConfig
        """
        return self._multi_version_config

    @multi_version_config.setter
    def multi_version_config(self, multi_version_config):
        """Sets the multi_version_config of this PropertyAdvancedConfig.


        :param multi_version_config: The multi_version_config of this PropertyAdvancedConfig.  # noqa: E501
        :type: MultiVersionConfig
        """

        self._multi_version_config = multi_version_config

    @property
    def mounted_concept_config(self):
        """Gets the mounted_concept_config of this PropertyAdvancedConfig.  # noqa: E501


        :return: The mounted_concept_config of this PropertyAdvancedConfig.  # noqa: E501
        :rtype: MountedConceptConfig
        """
        return self._mounted_concept_config

    @mounted_concept_config.setter
    def mounted_concept_config(self, mounted_concept_config):
        """Sets the mounted_concept_config of this PropertyAdvancedConfig.


        :param mounted_concept_config: The mounted_concept_config of this PropertyAdvancedConfig.  # noqa: E501
        :type: MountedConceptConfig
        """

        self._mounted_concept_config = mounted_concept_config

    @property
    def property_group(self):
        """Gets the property_group of this PropertyAdvancedConfig.  # noqa: E501


        :return: The property_group of this PropertyAdvancedConfig.  # noqa: E501
        :rtype: str
        """
        return self._property_group

    @property_group.setter
    def property_group(self, property_group):
        """Sets the property_group of this PropertyAdvancedConfig.


        :param property_group: The property_group of this PropertyAdvancedConfig.  # noqa: E501
        :type: str
        """
        allowed_values = ["TIME", "SUBJECT", "OBJECT", "LOC"]  # noqa: E501
        if (
            self.local_vars_configuration.client_side_validation
            and property_group not in allowed_values
        ):  # noqa: E501
            raise ValueError(
                "Invalid value for `property_group` ({0}), must be one of {1}".format(  # noqa: E501
                    property_group, allowed_values
                )
            )

        self._property_group = property_group

    @property
    def constraint(self):
        """Gets the constraint of this PropertyAdvancedConfig.  # noqa: E501


        :return: The constraint of this PropertyAdvancedConfig.  # noqa: E501
        :rtype: Constraint
        """
        return self._constraint

    @constraint.setter
    def constraint(self, constraint):
        """Sets the constraint of this PropertyAdvancedConfig.


        :param constraint: The constraint of this PropertyAdvancedConfig.  # noqa: E501
        :type: Constraint
        """

        self._constraint = constraint

    @property
    def sub_properties(self):
        """Gets the sub_properties of this PropertyAdvancedConfig.  # noqa: E501


        :return: The sub_properties of this PropertyAdvancedConfig.  # noqa: E501
        :rtype: list[SubProperty]
        """
        return self._sub_properties

    @sub_properties.setter
    def sub_properties(self, sub_properties):
        """Sets the sub_properties of this PropertyAdvancedConfig.


        :param sub_properties: The sub_properties of this PropertyAdvancedConfig.  # noqa: E501
        :type: list[SubProperty]
        """

        self._sub_properties = sub_properties

    @property
    def semantics(self):
        """Gets the semantics of this PropertyAdvancedConfig.  # noqa: E501


        :return: The semantics of this PropertyAdvancedConfig.  # noqa: E501
        :rtype: list[PredicateSemantic]
        """
        return self._semantics

    @semantics.setter
    def semantics(self, semantics):
        """Sets the semantics of this PropertyAdvancedConfig.


        :param semantics: The semantics of this PropertyAdvancedConfig.  # noqa: E501
        :type: list[PredicateSemantic]
        """

        self._semantics = semantics

    @property
    def logical_rule(self):
        """Gets the logical_rule of this PropertyAdvancedConfig.  # noqa: E501


        :return: The logical_rule of this PropertyAdvancedConfig.  # noqa: E501
        :rtype: LogicalRule
        """
        return self._logical_rule

    @logical_rule.setter
    def logical_rule(self, logical_rule):
        """Sets the logical_rule of this PropertyAdvancedConfig.


        :param logical_rule: The logical_rule of this PropertyAdvancedConfig.  # noqa: E501
        :type: LogicalRule
        """

        self._logical_rule = logical_rule

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(
                    map(
                        lambda item: (item[0], item[1].to_dict())
                        if hasattr(item[1], "to_dict")
                        else item,
                        value.items(),
                    )
                )
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, PropertyAdvancedConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, PropertyAdvancedConfig):
            return True

        return self.to_dict() != other.to_dict()



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
from abc import ABC
from typing import Type, List


class RESTable(ABC):
    """
    Abstract base class that can be serialized as REST model and submit to the SPG server.
    """

    @property
    def upstream_types(self) -> List[Type["RESTable"]]:
        """The types of upstream RESTable objects that the current RESTable object can support.

        Returns: RESTable type list.

        """
        return []

    @property
    def downstream_types(self) -> List[Type["RESTable"]]:
        """The types of downstream RESTable objects that the current RESTable object can support.

        Returns: RESTable type list.

        """
        return []

    def to_rest(self):
        """Convert a RESTable object to REST model that can be serialized.

        Returns: REST model.

        """
        raise NotImplementedError(
            f"`to_rest` is not currently supported for {self.__class__.__name__}."
        )

    @classmethod
    def from_rest(cls, rest_model):
        """Convert a REST model to RESTable object.

        Args:
            rest_model: REST model that needs to be converted to a RESTable object.

        Returns: Object inherits from RESTable.

        """
        raise NotImplementedError(
            f"`from_rest` is not currently supported for {cls.__name__}."
        )

    def submit(self):
        """Submit to the SPG Server.
        Usually, it is necessary to call the `to_rest` method firstly.

        """
        raise NotImplementedError(
            f"`submit` is not currently supported for {self.__class__.__name__}."
        )



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import sys
import click

from knext.reasoner.client import ReasonerClient


@click.option("--file", help="Path of DSL file.")
@click.option("--dsl", help="DSL string enclosed in double quotes.")
@click.option("--output", help="Output file.")
def execute_reasoner_job(file, dsl, output=None):
    """
    Submit asynchronous reasoner jobs to server by providing DSL file or string.
    """
    client = ReasonerClient()
    if file and not dsl:
        with open(file, "r") as f:
            dsl_content = f.read()
    elif not file and dsl:
        dsl_content = dsl
    else:
        click.secho("ERROR: Please choose either --file or --dsl.", fg="bright_red")
        sys.exit()
    client.execute(dsl_content, output)



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from enum import Enum
from typing import Dict, Type

from knext.builder.chain import BuilderChain


class AlterOperationEnum(str, Enum):
    Upsert = "UPSERT"
    Delete = "DELETE"


class BuilderJob:
    """Base class for all knowledge builder jobs.
    A builder job consists of multiple components.
    The declaration of components and the dependencies between components need to be implemented in `build` method.
    """

    parallelism: int = 1
    alter_operation: AlterOperationEnum = AlterOperationEnum.Upsert
    lead_to: bool

    _registry: Dict[str, Type] = {}
    _local_path: str
    _module_path: str
    _has_registered: bool = False

    def build(self) -> BuilderChain:
        """All classes as subclass of BuilderJob need to implement this method,
        to declare the DAG structure of the builder job.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} need to implement `build` method."
        )

    @classmethod
    def register(cls, name: str, local_path: str, module_path: str):
        """Register a class as subclass of BuilderJob with name and local_path.
        After registration, the subclass object can be inspected by `BuilderJob.by_name(job_name)`.
        """

        def add_subclass_to_registry(subclass: Type["BuilderJob"]):
            subclass.name = name
            subclass._local_path = local_path
            subclass._module_path = module_path
            if name in cls._registry:
                raise ValueError(
                    f"BuilderJob [{name}] conflict in {subclass._local_path} and {cls.by_name(name)._local_path}."
                )

            cls._registry[name] = subclass
            return subclass

        return add_subclass_to_registry

    @classmethod
    def by_name(cls, name: str):
        """Reflection from job name to subclass object of BuilderJob."""
        if name in BuilderJob._registry:
            subclass = BuilderJob._registry[name]
            return subclass
        else:
            raise ValueError(f"{name} is not a registered name for {cls.__name__}. ")

    @classmethod
    def has_registered(cls):
        return cls._has_registered



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.


import click

from knext.common.env import get_cfg_files, get_config


@click.option("--global", multiple=True)
@click.option("--local", multiple=True)
def edit_config(**kwargs):
    """
    Edit global or local configs.
    """
    global_cfg, global_cfg_path, local_cfg, local_cfg_path = get_cfg_files()
    for option, args in kwargs.items():
        if option == "global":
            if not global_cfg.has_section("global"):
                global_cfg.add_section("global")
            for arg in args:
                key, value = arg.split("=", 1)
                global_cfg.set("global", key, value)
                with open(global_cfg_path, "w") as config_file:
                    global_cfg.write(config_file)
        if option == "local":
            for arg in args:
                key, value = arg.split("=", 1)
                local_cfg.set("local", key, value)
                with open(local_cfg_path, "w") as config_file:
                    local_cfg.write(config_file)


def list_config():
    """
    List global and local knext configs.
    """
    config, _ = get_config()
    click.echo("[global]")
    if config.has_section("global"):
        for key, value in config.items("global"):
            click.echo(f"{key} = {value}")
    click.echo("[local]")
    if config.has_section("local"):
        for key, value in config.items("local"):
            click.echo(f"{key} = {value}")



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from typing import List, Dict

from pydantic import Field

from knext.builder import rest
from knext.common.base.runnable import Input, Output
from knext.builder.component.base import SourceReader


class CSVReader(SourceReader):
    """A source component that reading data from CSV file.

    Args:
        local_path: The local path of CSV file.
        columns: The column names that need to be read from the CSV file.
        start_row: The starting number of rows read from the CSV file.
                    If the CSV file includes a header, it needs to be greater than or equal to 2.
    Examples:
        source = CSVReader(
                    local_path="./builder/job/data/App.csv",
                    columns=["id", 'riskMark', 'useCert'],
                    start_row=2
                )
    """

    """The local path of CSV file."""
    local_path: str
    """The column names that need to be read from the CSV file."""
    columns: List[str]
    """The starting number of rows read from the CSV file.
    If the CSV file includes a header, it needs to be greater than or equal to 2."""
    start_row: int = Field(ge=1)

    @property
    def input_types(self) -> Input:
        return None

    @property
    def output_types(self) -> Output:
        return Dict[str, str]

    def invoke(self, input: Input):
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support being invoked separately."
        )

    def submit(self):
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support being submitted separately."
        )

    def to_rest(self):
        """Transforms `CSVReader` to REST model `CsvSourceNodeConfig`."""
        from pathlib import Path

        absolute_path = str(Path(self.local_path).resolve())
        config = rest.CsvSourceNodeConfig(
            start_row=self.start_row, url=absolute_path, columns=self.columns
        )
        return rest.Node(**super().to_dict(), node_config=config)

    @classmethod
    def from_rest(cls, node: rest.Node):
        return cls(
            local_path=node.node_config.url,
            columns=node.node_config.columns,
            start_row=node.node_config.start_row,
        )
